# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""debugger context viewer"""


from .stackviewer import StackViewer
from .threadsviewer import ThreadsViewer
from .variablesviewer import VariablesViewer
from ui.qt import Qt, QVBoxLayout, QWidget, QSplitter


class DebuggerContext(QWidget):

    """Implements the debugger context viewer"""

    def __init__(self, debugger, parent=None):
        QWidget.__init__(self, parent)
        self.__debugger = debugger
        self.__debugger.sigClientLine.connect(self.__onClientLine)
        self.__debugger.sigClientStack.connect(self.onClientStack)
        self.__debugger.sigClientThreadList.connect(self.__onClientThreadList)
        self.__debugger.sigClientVariables.connect(self.__onClientVariables)
        self.__debugger.sigClientVariable.connect(self.__onClientVariable)
        self.__debugger.sigClientThreadSet.connect(self.__onClientThreadSet)

        self.currentStack = None
        self.__createLayout()

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(1, 1, 1, 1)

        self.splitter = QSplitter(Qt.Vertical)

        self.variablesViewer = VariablesViewer(self.__debugger, self.splitter)
        self.stackViewer = StackViewer(self.__debugger, self.splitter)
        self.threadsViewer = ThreadsViewer(self.__debugger, self.splitter)

        self.splitter.addWidget(self.variablesViewer)
        self.splitter.addWidget(self.stackViewer)
        self.splitter.addWidget(self.threadsViewer)

        self.splitter.setCollapsible(0, False)
        self.splitter.setCollapsible(1, False)
        self.splitter.setCollapsible(2, False)

        verticalLayout.addWidget(self.splitter)

    def clear(self):
        """Clears everything"""
        self.variablesViewer.clear()
        self.stackViewer.clear()
        self.threadsViewer.clear()

    def __onClientLine(self, fileName, line, forStack):
        """Handles the signal from the debugged program"""
        del fileName    # unused argument
        del line        # unused argument
        if not forStack:
            self.__debugger.remoteThreadList()
            self.__debugger.remoteClientVariables(1, 0)  # globals
            self.__debugger.remoteClientVariables(0, 0)  # locals

    def onClientStack(self, stack):
        """Handles the signal from the debugged program"""
        self.stackViewer.populate(stack)

    def __onClientThreadList(self, currentThreadID, threadList):
        """Handles the thread list from the remote client"""
        self.threadsViewer.populate(currentThreadID, threadList)

    def __onClientVariables(self, scope, variables):
        """Handles the client variables lists"""
        frameNumber = self.stackViewer.getFrameNumber()
        if scope in [-1, 0, False]:
            # Empty list for local variables
            self.variablesViewer.updateVariables(False, frameNumber, variables)
        else:
            self.variablesViewer.updateVariables(True, frameNumber, variables)

    def __onClientVariable(self, scope, variables):
        """Handles the client variable list"""
        if scope in [-1, 0, False]:
            self.variablesViewer.updateVariable(False, variables)
        else:
            self.variablesViewer.updateVariable(True, variables)

    def __onClientThreadSet(self):
        """Handles the event of setting the current thread by the client"""
        self.__debugger.remoteClientVariables(1, 0)   # globals
        self.__debugger.remoteClientVariables(0, 0)   # locals

    def switchControl(self, isInIDE):
        """Switches the UI depending where the control flow is"""
        self.variablesViewer.switchControl(isInIDE)
        self.stackViewer.switchControl(isInIDE)
        self.threadsViewer.switchControl(isInIDE)

    def getCurrentFrameNumber(self):
        """Provides the current frame number"""
        return self.stackViewer.getFrameNumber()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Disassembling files and buffers"""

import sys
import dis
import os
import os.path
import py_compile
import marshal
import platform
import binascii
import struct
import time
from types import CodeType
from io import StringIO
from utils.fileutils import makeTempFile, saveToFile


OPT_NO_OPTIMIZATION = 0
OPT_OPTIMIZE_ASSERT = 1
OPT_OPTIMIZE_DOCSTRINGS = 2

_CONVERSION = {OPT_NO_OPTIMIZATION: 'no optimization',
               OPT_OPTIMIZE_ASSERT: 'assert optimization',
               OPT_OPTIMIZE_DOCSTRINGS: 'assert + docstring optimization'}

def optToString(optimization):
    """Converts optimization into a string"""
    try:
        return _CONVERSION[optimization]
    except KeyError:
        return 'unknown optimization'


def safeUnlink(path):
    """No exception unlink"""
    try:
        os.unlink(path)
    except:
        pass


def getCodeDisassembly(code):
    """Provides disassembly of a code object"""
    fileLikeObject = StringIO()
    dis.dis(code, file=fileLikeObject)
    fileLikeObject.seek(0)
    return fileLikeObject.read()


def recursiveDisassembly(codeObject, name=None):
    """Disassemble recursively"""
    what = 'module' if name is None else name
    res = '\n\nDisassembly of ' + what + ':\n' + getCodeDisassembly(codeObject)
    for item in codeObject.co_consts:
        if type(item) == CodeType:
            itemName = item.co_name
            if name:
                itemName = name + '.' + itemName
            res += recursiveDisassembly(item, itemName)
    return res


# The idea is taken from here:
# https://stackoverflow.com/questions/11141387/given-a-python-pyc-file-is-there-a-tool-that-let-me-view-the-bytecode
# https://stackoverflow.com/questions/32562163/how-can-i-understand-a-pyc-file-content
def getCompiledfileDisassembled(pycPath, pyPath, optimization,
                                forBuffer=False):
    """Reads the .pyc file and provides the plain text disassembly"""
    pycFile = open(pycPath, 'rb')

    magic = pycFile.read(4)
    timestamp = pycFile.read(4)

    size = None
    if sys.version_info.major == 3 and sys.version_info.minor >= 3:
        size = pycFile.read(4)
        size = struct.unpack('I', size)[0]

    code = marshal.load(pycFile)
    magic = binascii.hexlify(magic).decode('utf-8')
    timestamp = time.asctime(
        time.localtime(struct.unpack('I', b'D\xa5\xc2X')[0]))

    bufferSpec = ''
    if forBuffer:
        bufferSpec = ' (unsaved buffer)'
    return '\n'.join(
        ['-' * 80,
         'Python version: ' + platform.python_version(),
         'Python interpreter path: ' + sys.executable,
         'Interpreter magic: ' + magic,
         'Interpreter timestamp: ' + timestamp,
         'Python module: ' + pyPath + bufferSpec,
         'Optimization: ' + optToString(optimization),
         'Code size: ' + str(size),
         '-' * 80,
         recursiveDisassembly(code)])


def getFileDisassembled(path, optimization):
    """Dsassembles a file"""
    if not os.path.exists(path):
        raise Exception("Cannot find " + path + " to disassemble")
    if not os.access(path, os.R_OK):
        raise Exception("No read permissions for " + path)

    tempPycFile = makeTempFile(suffix='.pyc')
    try:
        py_compile.compile(path, tempPycFile,
                           doraise=True, optimize=optimization)
    except Exception as exc:
        safeUnlink(tempPycFile)
        raise Exception("Cannot disassemble file " + path + ': ' + str(exc))

    try:
        result = getCompiledfileDisassembled(tempPycFile, path, optimization)
    except:
        safeUnlink(tempPycFile)
        raise

    safeUnlink(tempPycFile)
    return result


def getBufferDisassembled(content, encoding, path, optimization):
    """Disassembles a memory buffer"""
    tempSrcFile = makeTempFile(suffix='.py')
    tempPycFile = makeTempFile(suffix='.pyc')

    try:
        saveToFile(tempSrcFile, content, allowException=True, enc=encoding)
        py_compile.compile(tempSrcFile, tempPycFile, path,
                           doraise=True, optimize=optimization)
    except Exception:
        safeUnlink(tempSrcFile)
        safeUnlink(tempPycFile)
        raise

    try:
        result = getCompiledfileDisassembled(tempPycFile, path,
                                             optimization, True)
    except:
        safeUnlink(tempSrcFile)
        safeUnlink(tempPycFile)
        raise

    safeUnlink(tempSrcFile)
    safeUnlink(tempPycFile)
    if path:
        return result.replace('file "' + path + '",',
                              'unsaved buffer "' + path + '",')
    return result



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""file related utils"""

# pylint: disable=W0603
# pylint: disable=W0702
# pylint: disable=W0703


import os
from os.path import (islink, exists, split, join, sep, basename, realpath,
                     normpath, isabs, dirname, isdir)
import logging
import json
from errno import EACCES, ENOENT
import tempfile
import magic

# Qutepart has a few maps which help to map a file to a syntax.
from qutepart import Qutepart

from ui.qt import QImageReader
from .pixmapcache import getIcon
from .config import DEFAULT_ENCODING


__QTSupportedImageFormats = [fmt.data().decode(DEFAULT_ENCODING) for fmt in
                             QImageReader.supportedImageFormats()]
VIEWABLE_IMAGE_MIMES = ['image/' + ext for ext in __QTSupportedImageFormats]


# NOTES
# The standard set of mime types does not cover all the required cases.
# In particular:
# - makefiles are not reliably distinguished
# - code and header files are detected as the same type but the icon needs to
#   be different
# - broken symbolic links are not distinguished


def __getXmlSyntaxFile(fName):
    """Checks the Qutepart mapping of a file extension to a syntax file.

    Returns: None if not found or a syntax file name
    """
    # Special cases: codimension project files
    #                makefiles
    if fName.endswith('.cdm'):
        return 'ini.xml'
    if fName.endswith('.cdm3'):
        return 'json.xml'
    if fName.endswith('.ts'):
        return 'xml.xml'
    if 'makefile' in fName.lower():
        return "makefile.xml"

    for regExp, xmlFileName in \
                Qutepart._globalSyntaxManager._extensionToXmlFileName.items():
        if regExp.match(fName):
            return xmlFileName
    return None


def __getMimeByXmlSyntaxFile(xmlSyntaxFile):
    """Checks the Qutepart mapping of a mime type to a syntax file.

    Returns a mime type or None
    """
    candidates = []
    for mime, xmlFileName in \
                Qutepart._globalSyntaxManager._mimeTypeToXmlFileName.items():
        if xmlFileName == xmlSyntaxFile:
            candidates.append(mime)
    if not candidates:
        # The qutepart syntax DB misses a markdown mime
        if xmlSyntaxFile == 'markdown.xml':
            return 'text/x-markdown'
        # ... and mako mime
        if xmlSyntaxFile == 'mako.xml':
            return 'text/x-mako'
        return None
    if len(candidates) == 1:
        return candidates[0]

    # Prefer the text candidate
    for item in candidates:
        if item.startswith('text'):
            return item
    return candidates[0]


def getXmlSyntaxFileByMime(mime):
    """Checks the Qutepart mapping of a mime type to a syntax file.

    Returns an xml syntax file or None
    """
    try:
        return Qutepart._globalSyntaxManager._mimeTypeToXmlFileName[mime]
    except KeyError:
        if mime == 'text/x-c++':
            return 'cpp.xml'
        return None


__magic = magic.Magic(mime=True)


def isFileSearchable(fName, checkForBrokenLink=True):
    """Returns True if it makes sense to search for text in that file"""
    mime, _, syntaxFile = getFileProperties(fName, checkForBrokenLink,
                                            skipCache=False)
    if syntaxFile is not None:
        # The files like libQt5Widgets.so.5 are mistakenly detected
        # as a man doc files (mandoc.xml)
        if mime:
            if mime.endswith('x-sharedlib'):
                return False
        return True
    if mime is None:
        return False
    return mime.startswith('text/')


def isImageViewable(mime):
    """True if QT can show the image"""
    if mime is None:
        return False
    return mime in VIEWABLE_IMAGE_MIMES


def isImageFile(fName):
    """True is the file is a viewable image"""
    mime, _, _ = getFileProperties(fName, checkForBrokenLink=True,
                                   skipCache=False)
    return isImageViewable(mime)


def isFileOpenable(fName):
    """True if codimension can open the file"""
    mime, _, syntaxFile = getFileProperties(fName, True, False)
    if syntaxFile is not None:
        return True
    return isImageViewable(mime)


__syntaxToIcon = None


def __initSyntaxToIcon():
    """Prevents calling getIcon before QApplication is created"""
    global __syntaxToIcon
    __syntaxToIcon = {
        'stata.xml': getIcon('filemisc.png'),
        'sisu.xml': getIcon('filemisc.png'),
        'mandoc.xml': getIcon('fileman.png'),
        'fgl-4gl.xml': getIcon('filemisc.png'),
        'ansforth94.xml': getIcon('filemisc.png'),
        'abap.xml': getIcon('filemisc.png'),
        'abc.xml': getIcon('filemisc.png'),
        'asm-avr.xml': getIcon('filemisc.png'),
        'freebasic.xml': getIcon('filemisc.png'),
        'wml.xml': getIcon('filemisc.png'),
        'd.xml': getIcon('filed.png'),
        'sql-oracle.xml': getIcon('filemisc.png'),
        'maxima.xml': getIcon('filemisc.png'),
        'fortran.xml': getIcon('filemisc.png'),
        'gdl.xml': getIcon('filemisc.png'),
        'haxe.xml': getIcon('filemisc.png'),
        'asm-m68k.xml': getIcon('filemisc.png'),
        'j.xml': getIcon('filemisc.png'),
        'lilypond.xml': getIcon('filemisc.png'),
        'asm-dsp56k.xml': getIcon('filemisc.png'),
        'jsp.xml': getIcon('filemisc.png'),
        'objectivecpp.xml': getIcon('filemisc.png'),
        'mab.xml': getIcon('filemisc.png'),
        'fgl-per.xml': getIcon('filemisc.png'),
        'pgn.xml': getIcon('filemisc.png'),
        'picsrc.xml': getIcon('filemisc.png'),
        'perl.xml': getIcon('fileperl.png'),
        'xharbour.xml': getIcon('filemisc.png'),
        'r.xml': getIcon('filemisc.png'),
        'rmarkdown.xml': getIcon('filemisc.png'),
        'relaxng.xml': getIcon('filemisc.png'),
        'verilog.xml': getIcon('filemisc.png'),
        'ada.xml': getIcon('fileada.png'),
        'hunspell-aff.xml': getIcon('filemisc.png'),
        'agda.xml': getIcon('filemisc.png'),
        'ahdl.xml': getIcon('filemisc.png'),
        'ahk.xml': getIcon('filemisc.png'),
        'postscript.xml': getIcon('filepostscript.png'),
        'ample.xml': getIcon('filemisc.png'),
        'ansys.xml': getIcon('filemisc.png'),
        'actionscript.xml': getIcon('filemisc.png'),
        'asn1.xml': getIcon('filemisc.png'),
        'asp.xml': getIcon('filemisc.png'),
        'awk.xml': getIcon('fileawk.png'),
        'bash.xml': getIcon('fileshell.png'),
        'dosbat.xml': getIcon('filemisc.png'),
        'latex.xml': getIcon('filetex.png'),
        'bibtex.xml': getIcon('filetex.png'),
        'boo.xml': getIcon('filemisc.png'),
        'component-pascal.xml': getIcon('filepascal.png'),
        'gdb.xml': getIcon('filemisc.png'),
        '4dos.xml': getIcon('filemisc.png'),
        'c.xml': getIcon('filec.png'),
        'ccss.xml': getIcon('filecss.png'),
        'coldfusion.xml': getIcon('filemisc.png'),
        'nagios.xml': getIcon('filemisc.png'),
        'cg.xml': getIcon('filemisc.png'),
        'cgis.xml': getIcon('filemisc.png'),
        'chicken.xml': getIcon('filemisc.png'),
        'haskell.xml': getIcon('filemisc.png'),
        'cisco.xml': getIcon('filemisc.png'),
        'opencl.xml': getIcon('filemisc.png'),
        'clojure.xml': getIcon('filemisc.png'),
        'cmake.xml': getIcon('filemake.png'),
        'coffee.xml': getIcon('filemisc.png'),
        'logtalk.xml': getIcon('filemisc.png'),
        'crk.xml': getIcon('filemisc.png'),
        'cs.xml': getIcon('filemisc.png'),
        'tcsh.xml': getIcon('fileshell.png'),
        'css.xml': getIcon('filecss.png'),
        'context.xml': getIcon('filemisc.png'),
        'cue.xml': getIcon('filemisc.png'),
        'curry.xml': getIcon('filemisc.png'),
        'xml.xml': getIcon('filexml.png'),
        'hunspell-dat.xml': getIcon('filemisc.png'),
        'prolog.xml': getIcon('filemisc.png'),
        'modula-2.xml': getIcon('filemisc.png'),
        'desktop.xml': getIcon('filemisc.png'),
        'hunspell-dic.xml': getIcon('filemisc.png'),
        'diff.xml': getIcon('filediff.png'),
        'dot.xml': getIcon('filemisc.png'),
        'doxygen.xml': getIcon('filemisc.png'),
        'dtd.xml': getIcon('filemisc.png'),
        'euphoria.xml': getIcon('filemisc.png'),
        'email.xml': getIcon('fileemail.png'),
        'erlang.xml': getIcon('filemisc.png'),
        'fasm.xml': getIcon('filemisc.png'),
        'ferite.xml': getIcon('filemisc.png'),
        'lex.xml': getIcon('filemisc.png'),
        'glsl.xml': getIcon('filemisc.png'),
        'fsharp.xml': getIcon('filemisc.png'),
        'ftl.xml': getIcon('filemisc.png'),
        'grammar.xml': getIcon('filemisc.png'),
        'gap.xml': getIcon('filemisc.png'),
        'glosstex.xml': getIcon('filetex.png'),
        'ruby.xml': getIcon('filemisc.png'),
        'gnuplot.xml': getIcon('filemisc.png'),
        'go.xml': getIcon('filemisc.png'),
        'groovy.xml': getIcon('filemisc.png'),
        'scheme.xml': getIcon('filemisc.png'),
        'haml.xml': getIcon('filemisc.png'),
        'hamlet.xml': getIcon('filemisc.png'),
        'spice.xml': getIcon('filemisc.png'),
        'html.xml': getIcon('filehtml.png'),
        'rhtml.xml': getIcon('filemisc.png'),
        'progress.xml': getIcon('filemisc.png'),
        'vcard.xml': getIcon('filevcard.png'),
        'idl.xml': getIcon('fileidl.png'),
        'hunspell-idx.xml': getIcon('filemisc.png'),
        'bmethod.xml': getIcon('filemisc.png'),
        'opal.xml': getIcon('filemisc.png'),
        'html-php.xml': getIcon('filemisc.png'),
        'inform.xml': getIcon('filemisc.png'),
        'ini.xml': getIcon('fileprops.png'),
        'jam.xml': getIcon('filemisc.png'),
        'java.xml': getIcon('filejava.png'),
        'jira.xml': getIcon('filemisc.png'),
        'julia.xml': getIcon('filemisc.png'),
        'javascript.xml': getIcon('filejs.png'),
        'json.xml': getIcon('filemisc.png'),
        'k.xml': getIcon('filemisc.png'),
        'kbasic.xml': getIcon('filemisc.png'),
        'literate-curry.xml': getIcon('filemisc.png'),
        'ld.xml': getIcon('filemisc.png'),
        'ldif.xml': getIcon('filemisc.png'),
        'less.xml': getIcon('filemisc.png'),
        'literate-haskell.xml': getIcon('filemisc.png'),
        'commonlisp.xml': getIcon('filemisc.png'),
        'lsl.xml': getIcon('filemisc.png'),
        'lua.xml': getIcon('filemisc.png'),
        'octave.xml': getIcon('filemisc.png'),
        'm3u.xml': getIcon('filemisc.png'),
        'm4.xml': getIcon('filemisc.png'),
        'mako.xml': getIcon('filemisc.png'),
        'mediawiki.xml': getIcon('filemisc.png'),
        'mel.xml': getIcon('filemisc.png'),
        'metafont.xml': getIcon('filemisc.png'),
        'makefile.xml': getIcon('filemake.png'),
        'ocaml.xml': getIcon('filemisc.png'),
        'ocamllex.xml': getIcon('filemisc.png'),
        'ocamlyacc.xml': getIcon('filemisc.png'),
        'modelica.xml': getIcon('filemisc.png'),
        'carto-css.xml': getIcon('filecss.png'),
        'mergetagtext.xml': getIcon('filemisc.png'),
        'mup.xml': getIcon('filemisc.png'),
        'nemerle.xml': getIcon('filemisc.png'),
        'mathematica.xml': getIcon('filemisc.png'),
        'nesc.xml': getIcon('filemisc.png'),
        'nsis.xml': getIcon('filemisc.png'),
        'noweb.xml': getIcon('filemisc.png'),
        'lpc.xml': getIcon('filemisc.png'),
        'oors.xml': getIcon('filemisc.png'),
        'pascal.xml': getIcon('filepascal.png'),
        'purebasic.xml': getIcon('filemisc.png'),
        'pig.xml': getIcon('filemisc.png'),
        'pike.xml': getIcon('filemisc.png'),
        'pli.xml': getIcon('filemisc.png'),
        'gettext.xml': getIcon('filemisc.png'),
        'povray.xml': getIcon('filemisc.png'),
        'puppet.xml': getIcon('filemisc.png'),
        'ppd.xml': getIcon('filemisc.png'),
        'qmake.xml': getIcon('filemake.png'),
        'protobuf.xml': getIcon('filemisc.png'),
        'python.xml': getIcon('filepython.png'),
        'qml.xml': getIcon('filemisc.png'),
        'winehq.xml': getIcon('filemisc.png'),
        'replicode.xml': getIcon('filemisc.png'),
        'rexx.xml': getIcon('filemisc.png'),
        'rib.xml': getIcon('filemisc.png'),
        'relaxngcompact.xml': getIcon('filemisc.png'),
        'rapidq.xml': getIcon('filemisc.png'),
        'rust.xml': getIcon('filemisc.png'),
        'rest.xml': getIcon('filemisc.png'),
        'rtf.xml': getIcon('filertf.png'),
        'sather.xml': getIcon('filemisc.png'),
        'scala.xml': getIcon('filemisc.png'),
        'sci.xml': getIcon('filemisc.png'),
        'scss.xml': getIcon('filecss.png'),
        'sed.xml': getIcon('filemisc.png'),
        'sgml.xml': getIcon('filemisc.png'),
        'zsh.xml': getIcon('fileshell.png'),
        'sieve.xml': getIcon('filemisc.png'),
        'sml.xml': getIcon('filemisc.png'),
        'rpmspec.xml': getIcon('filemisc.png'),
        'valgrind-suppression.xml': getIcon('filemisc.png'),
        'systemverilog.xml': getIcon('filemisc.png'),
        'tads3.xml': getIcon('filemisc.png'),
        'txt2tags.xml': getIcon('filemisc.png'),
        'tcl.xml': getIcon('filetcl.png'),
        'texinfo.xml': getIcon('filetex.png'),
        'textile.xml': getIcon('filemisc.png'),
        'taskjuggler.xml': getIcon('filemisc.png'),
        'toml.xml': getIcon('filemisc.png'),
        'template-toolkit.xml': getIcon('filemisc.png'),
        'uscript.xml': getIcon('filemisc.png'),
        'vala.xml': getIcon('filemisc.png'),
        'monobasic.xml': getIcon('filemisc.png'),
        'varnishcc4.xml': getIcon('filemisc.png'),
        'varnish4.xml': getIcon('filemisc.png'),
        'vhdl.xml': getIcon('filemisc.png'),
        'velocity.xml': getIcon('filemisc.png'),
        'vera.xml': getIcon('filemisc.png'),
        'varnishtest4.xml': getIcon('filemisc.png'),
        'vrml.xml': getIcon('filemisc.png'),
        'xslt.xml': getIcon('filexml.png'),
        'xul.xml': getIcon('filemisc.png'),
        'yaml.xml': getIcon('filemisc.png'),
        'yacas.xml': getIcon('filemisc.png'),
        'yacc.xml': getIcon('filemisc.png'),
        'zonnon.xml': getIcon('filemisc.png'),
        'asterisk.xml': getIcon('filemisc.png'),
        'git-ignore.xml': getIcon('filemisc.png'),
        'kdesrc-buildrc.xml': getIcon('filemisc.png'),
        'changelog.xml': getIcon('filemisc.png'),
        'dockerfile.xml': getIcon('filemisc.png'),
        'kconfig.xml': getIcon('filemisc.png'),
        'ilerpg.xml': getIcon('filemisc.png'),
        'apache.xml': getIcon('filemisc.png'),
        'debiancontrol.xml': getIcon('filemisc.png'),
        'fstab.xml': getIcon('filemisc.png'),
        'git-rebase.xml': getIcon('filemisc.png'),
        'gitolite.xml': getIcon('filemisc.png'),
        'meson.xml': getIcon('filemisc.png'),
        'xorg.xml': getIcon('filemisc.png'),
        'markdown.xml': getIcon('filemarkdown.png')}


# Some specific cases for various binaries to be mapped to a certain icon
__mimeToIcon = None


def __initMimeToIcon():
    """Prevents calling getIcon() before QApplication is created"""
    global __mimeToIcon
    __mimeToIcon = {
        'application/x-executable': getIcon('filebinary.png'),
        'application/x-sharedlib': getIcon('fileso.png'),
        'application/x-coredump': getIcon('filemisc.png'),
        'application/pdf': getIcon('filepdf.png'),
        'application/x-bzip2': getIcon('filearchive.png'),
        'application/x-gzip': getIcon('filearchive.png'),
        'application/zip': getIcon('filearchive.png'),
        'application/x-xz': getIcon('filearchive.png'),
        'application/x-rar': getIcon('filearchive.png'),
        'application/x-tar': getIcon('filetar.png'),
        'application/x-object': getIcon('fileso.png')}


def __getIcon(xmlSyntaxFile, mime, fBaseName):
    """Provides an icon for a file"""
    if '.' in fBaseName:
        fileExtension = fBaseName.split('.')[-1].lower()
    else:
        fileExtension = ''

    if xmlSyntaxFile is not None:
        # There are a few special cases:
        # - svg is detected as xml.xml
        # - c/c++/h/h++ are detected as cpp.xml
        # - cdm project is detected as ini.xml
        if xmlSyntaxFile == 'ini.xml':
            if fileExtension == 'cdm':
                return getIcon('fileproject.png')
        if xmlSyntaxFile == 'json.xml':
            if fileExtension == 'cdm3':
                return getIcon('fileproject.png')
        if xmlSyntaxFile == 'cpp.xml':
            if 'h' in fileExtension:
                if fBaseName.endswith('.ui.h'):
                    return getIcon('filedesigner.png')
                if fBaseName.endswith('.h'):    # Capital H is for C++
                    return getIcon('filecheader.png')
                return getIcon('filecppheader.png')
            return getIcon('filecpp.png')
        if xmlSyntaxFile == 'xml.xml':
            if fileExtension == 'svg':
                return getIcon('filesvg.png')
            if fileExtension == "ui":
                return getIcon('filedesigner.png')
            if fileExtension in ['ts', 'qm']:
                return getIcon('filelinguist2.png')
            if fileExtension == 'qrc':
                return getIcon('fileresource.png')

        try:
            if __syntaxToIcon is None:
                __initSyntaxToIcon()
            return __syntaxToIcon[xmlSyntaxFile]
        except KeyError:
            return getIcon('filemisc.png')

    # Here: no luck with a syntax file
    # This could be an image or compiled python or binary of some sort...
    # Or a text which does not need a syntax highlight

    if fileExtension in __QTSupportedImageFormats:
        return getIcon('filepixmap.png')

    try:
        if mime is not None:
            if __mimeToIcon is None:
                __initMimeToIcon()
            return __mimeToIcon[mime]
    except KeyError:
        pass

    if fileExtension in ['pyc', 'pyo']:
        return getIcon('filepythoncompiled.png')

    if mime is not None:
        if mime == 'application/x-executable':
            return getIcon('filebinary.png')
        if mime.startswith('text/'):
            return getIcon('filetext.png')
    return getIcon('filemisc.png')


def __getMagicMime(fName):
    """Uses the magic module to retrieve the file mime.

    The mime could be None if e.g. the file does not exist.
    The second bool tells if it was a permission denied case.
    """
    try:
        # E.g.: 'text/x-shellscript'
        output = __magic.from_file(realpath(fName))
        return output, False
    except OSError as exc:
        if exc.errno == EACCES:
            return None, True
        if exc.errno == ENOENT:
            return None, False
        logging.error(str(exc))
        return None, False
    except Exception as exc:
        # Most probably the file does not exist
        logging.error(str(exc))
        return None, False


def getMagicMimeFromBuffer(txt):
    """Guesses the buffer text mime type"""
    try:
        return __magic.from_buffer(txt)
    except:
        return None


# Cache is initialized with:
# - an unknown file type properties: ''
# - a directory: '/'
# - broken symlink: '.'
__filePropertiesCache = None


def __initFilePropertiesCache():
    """Prevents calling getIcon before a QApplication is created"""
    global __filePropertiesCache
    __filePropertiesCache = {
        '': [None, getIcon('filemisc.png'), None],
        '/': ['inode/directory', getIcon('dirclosed.png'), None],
        '.': ['inode/broken-symlink', getIcon('filebrokenlink.png'), None]}


def getFileProperties(fName, checkForBrokenLink=True, skipCache=False):
    """Provides the following properties:

    - mime type (could be None)
    - icon
    - syntax file name (could be None)

    Works for non-existing files too.
    Special cases:
    - fName ends with os.path.sep => directory
    - fName is empy or None => unknown file type
    """
    global __filePropertiesCache
    if __filePropertiesCache is None:
        __initFilePropertiesCache()

    if not fName:
        return __filePropertiesCache['']

    if fName.endswith(sep):
        return __filePropertiesCache['/']

    if checkForBrokenLink and islink(fName):
        if not exists(fName):
            return __filePropertiesCache['.']

    if not skipCache and fName in __filePropertiesCache:
        value = __filePropertiesCache[fName]
        if value[0] is None:
            mime, _ = __getMagicMime(fName)
            if mime is not None:
                value[0] = mime
                __filePropertiesCache[fName] = value
        return value

    if skipCache:
        # Remove from cache so that the old value does not stuck there is case
        # if e.g. the file became empty
        __filePropertiesCache.pop(fName, None)

    # The function should work both for existing and non-existing files
    try:
        # If a file exists then it could be a symbolic link to
        # a different name file
        fBaseName = basename(realpath(fName))
    except:
        # File may not exist
        fBaseName = basename(fName)

    if '.' in fBaseName:
        fileExtension = fBaseName.split('.')[-1].lower()
    else:
        fileExtension = ''

    syntaxFile = __getXmlSyntaxFile(fBaseName)
    if syntaxFile is None:
        denied = False
        # Special case: this could be a QT supported image
        if fileExtension in __QTSupportedImageFormats:
            mime = 'image/' + fileExtension
        elif 'readme' in fBaseName.lower():
            mime = 'text/plain'
        else:
            mime, denied = __getMagicMime(fName)
            if mime == 'inode/x-empty':
                # Special case for the empty files; treat them as text files
                # till they are changed, i.e. do not memorize them in cache.
                # The magic library detects them as binary (without extensions)
                return [mime, getIcon('filemisc.png'),
                        getXmlSyntaxFileByMime('text/plain')]
            if mime is not None:
                syntaxFile = getXmlSyntaxFileByMime(mime)

        cacheValue = [mime,
                      getIcon('filedenied.png') if denied else
                      __getIcon(syntaxFile, mime, fBaseName),
                      syntaxFile]
        __filePropertiesCache[fName] = cacheValue
        return cacheValue

    # syntax file was successfully identified.
    # Detect the mime type by a syntax file
    if fileExtension == 'cdm':
        mime = 'text/x-codimension'
    elif fileExtension == 'cdm3':
        mime = 'text/x-codimension3'
    else:
        mime = __getMimeByXmlSyntaxFile(syntaxFile)
        if mime is None:
            mime, _ = __getMagicMime(fName)

    if fileExtension == 'o' and syntaxFile == 'lpc.xml':
        # lpc.xml is bound to .o extension i.e. exactly object files!
        if 'object' in mime:
            syntaxFile = None
    if fileExtension == 'a' and syntaxFile == 'ada.xml':
        tryMime, _ = __getMagicMime(fName)
        if 'x-archive' in tryMime:
            mime = tryMime
            syntaxFile = None
    if fileExtension == 'ttf' and syntaxFile == 'template-toolkit.xml':
        tryMime, _ = __getMagicMime(fName)
        if 'font' in tryMime:
            mime = tryMime
            syntaxFile = None
    if fileExtension == 'dat' and syntaxFile == 'hunspell-dat.xml':
        tryMime, _ = __getMagicMime(fName)
        if 'octet-stream' in tryMime:
            mime = tryMime
            syntaxFile = None

    cacheValue = [mime, __getIcon(syntaxFile, mime, fBaseName), syntaxFile]
    __filePropertiesCache[fName] = cacheValue
    return cacheValue


def compactPath(path, width, measure=len):
    """Provides a compacted path fitting inside the given width.

    measure - ref to a function used to get the length of the string
    """
    if measure(path) <= width:
        return path

    dots = '...'

    head, tail = split(path)
    mid = len(head) // 2
    head1 = head[:mid]
    head2 = head[mid:]

    while head1:
        path = join("%s%s%s" % (head1, dots, head2), tail)
        if measure(path) <= width:
            return path
        head1 = head1[:-1]
        head2 = head2[1:]

    path = join(dots, tail)
    if measure(path) <= width:
        return path

    while tail:
        path = "%s%s" % (dots, tail)
        if measure(path) <= width:
            return path
        tail = tail[1:]
    return ''


def isPythonMime(mime):
    """True if it is a python mime"""
    if mime is None:
        return False
    return 'python' in mime


def isMarkdownMime(mime):
    """True if it is a markdown mime"""
    if mime:
        return 'markdown' in mime
    return False


def isPythonFile(fName):
    """True if it is a python file"""
    mime, _, _ = getFileProperties(fName)
    return isPythonMime(mime)


def isPythonCompiledFile(fName):
    """True if it is a python compiled file"""
    mime, _, _ = getFileProperties(fName)
    if mime is None:
        return False
    if 'octet-stream' in mime:
        if fName.endswith('.pyc') or fName.endswith('.pyo'):
            return True
    return False


def isCDMProjectMime(mime):
    """True if it is a codimension project mime"""
    if mime is None:
        return False
    return 'x-codimension3' in mime


def isCDMProjectFile(fName):
    """True if it is a codimension project file"""
    mime, _, _ = getFileProperties(fName)
    if mime is None:
        return False
    return 'x-codimension3' in mime


# Utility functions to save/load generic JSON
def loadJSON(fileName, errorWhat, defaultValue):
    """Generic JSON loading"""
    try:
        with open(fileName, 'r', encoding=DEFAULT_ENCODING) as diskfile:
            return json.load(diskfile)
    except Exception as exc:
        logging.error('Error loading %s (from %s): %s',
                      errorWhat, fileName, str(exc))
        return defaultValue


def saveJSON(fileName, values, errorWhat):
    """Generic JSON saving"""
    try:
        with open(fileName, 'w', encoding=DEFAULT_ENCODING) as diskfile:
            json.dump(values, diskfile, indent=4)
    except Exception as exc:
        logging.error('Error saving %s (to %s): %s',
                      errorWhat, fileName, str(exc))
        return False
    return True


def getFileContent(fileName, allowException=True, enc=DEFAULT_ENCODING):
    """Provides the file content"""
    try:
        with open(fileName, 'r', encoding=enc) as diskfile:
            content = diskfile.read()
        return content
    except Exception as exc:
        if allowException:
            raise
        logging.error('Error reading from file %s: %s', fileName, str(exc))
        return None


def saveToFile(fileName, content, allowException=True, enc=DEFAULT_ENCODING):
    """Overwrites the file content"""
    try:
        with open(fileName, 'w', encoding=enc) as diskfile:
            diskfile.write(content)
        return True
    except Exception as exc:
        if allowException:
            raise
        logging.error('Error writing to file %s: %s', fileName, str(exc))
    return False


def saveBinaryToFile(fileName, content, allowException=True):
    """Overwrites the file content"""
    try:
        with open(fileName, 'wb') as diskfile:
            diskfile.write(content)
        return True
    except Exception as exc:
        if allowException:
            raise
        logging.error('Error writing to file %s: %s', fileName, str(exc))
    return False


def makeTempFile(prefix='cdm_', suffix=None):
    """Creates a temporary file"""
    fileHandle, fileName = tempfile.mkstemp(suffix, prefix)
    os.close(fileHandle)
    return fileName


def resolveLink(path):
    """Resolves links and detects loops"""
    paths_seen = []
    while islink(path):
        if path in paths_seen:
            # Already seen this path, so we must have a symlink loop
            return path, True
        paths_seen.append(path)
        # Resolve where the link points to
        resolved = os.readlink(path)
        if not isabs(resolved):
            dir_name = dirname(path)
            path = normpath(dir_name + sep + resolved)
        else:
            path = normpath(resolved)
    return path, False


def isCreatable(path):
    """Checks if the path is creatable"""
    if not isabs(path):
        raise Exception('The isCreatable() function '
                        'supports absolute paths only')
    if exists(path):
        raise Exception('The isCreatable() function '
                        'is only for not existing items')

    parts = path[1:].split(sep)
    for count in range(len(parts) - 1, 0, -1):
        # count is the number of path items to be used for a check
        candidatePath = sep + sep.join(parts[0:count])
        if exists(candidatePath):
            if isdir(candidatePath):
                if os.access(candidatePath, os.W_OK):
                    return True, None
                return False, 'No write permissions in ' + candidatePath

            return False, 'The path part ' + candidatePath + \
                          ' points to a non directory item'

    # Check the root directory
    if os.access('/', os.W_OK):
        return True, None

    return False, 'No write permissions to the root directory'



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Utilities to build completeion lists to suggest to the user"""


import os
import sys
import logging
import jedi
from utils.globals import GlobalData
from utils.fileutils import getFileContent
from .bufferutils import getEditorTags


def getCallSignatures(editor, fileName):
    """Provides a list of call signatures"""
    line, pos = editor.cursorPosition
    try:
        script = getJediScript(editor.text, line + 1, pos,
                               fileName if fileName else '')
        return script.call_signatures()
    except Exception as exc:
        logging.error('jedi library failed to provide call signatures: ' +
                      str(exc))
    return []


def getDefinitions(editor, fileName):
    """Provides the definition location or None"""
    result = []
    line, pos = editor.cursorPosition

    try:
        script = getJediScript(editor.text, line + 1, pos,
                               fileName if fileName else '')
        definitions = script.goto_definitions()

        # Filter out those on which there is no way to jump
        for definition in definitions:
            path = definition.module_path
            if definition.line is None or definition.column is None:
                continue
            if path is None and definition.in_builtin_module():
                continue
            result.append([path if path else fileName,
                           definition.line, definition.column,
                           definition.type, definition.docstring(),
                           definition.module_name])
    except Exception as exc:
        logging.error('jedi library failed to provide definitions: ' +
                      str(exc))
    return result


def getOccurrences(editor, fileName, line=None, pos=None):
    """Provides occurences for the current editor position"""
    if editor is not None:
        line, pos = editor.cursorPosition
        line += 1
        text = editor.text
    else:
        if pos > 0:
            pos -= 1
        try:
            text = getFileContent(fileName)
        except Exception as exc:
            logging.error('Cannot read file ' + fileName + ': ' + str(exc))
            return []

    try:
        script = getJediScript(text, line, pos,
                               fileName if fileName else '', False)
        return script.usages()
    except Exception as exc:
        logging.error('jedi library failed to provide usages: ' +
                      str(exc))
    return []


def getJediScript(source, line, column, srcPath, needSysPath=True):
    """Provides the jedi Script object considering the current project"""
    jedi.settings.additional_dynamic_modules = []
    paths = GlobalData().originalSysPath[:] if needSysPath else []

    # This make relative imports resolvable
    if os.path.isabs(srcPath):
        dirName = os.path.dirname(srcPath)
        if dirName not in paths:
            paths.append(dirName)

    project = GlobalData().project
    if not project.isLoaded():
        # Add the other opened files if so
        mainWindow = GlobalData().mainWindow
        for path in mainWindow.editorsManager().getOpenedList():
            if path[0]:
                if path[0].lower().endswith('.py'):
                    jedi.settings.additional_dynamic_modules.append(path[0])
        return jedi.Script(source, line, column, srcPath, sys_path=paths)

    for path in project.getImportDirsAsAbsolutePaths():
        if path not in paths:
            paths.append(path)
    projectDir = project.getProjectDir()
    if projectDir not in paths:
        paths.append(projectDir)

    return jedi.Script(source, line, column, srcPath, sys_path=paths)


def getCompletionList(editor, fileName):
    """Provides a list for completion"""
    if not editor.isPythonBuffer():
        return list(getEditorTags(editor))

    items = []
    line, pos = editor.cursorPosition
    try:
        script = getJediScript(editor.text, line + 1, pos,
                               fileName if fileName else '')

        for item in script.completions():
            items.append(item.name)
    except Exception as exc:
        logging.error('jedi library could not provide completions: ' +
                      str(exc))

    if items:
        return items

    # jedi provided nothing => last resort: words in the editor buffer
    return list(getEditorTags(editor))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Pyflakes results viewer"""

from html import escape
from radon.complexity import cc_rank
from utils.pixmapcache import getIcon, getPixmap
from utils.fileutils import isPythonMime, isPythonFile
from analysis.ierrors import getBufferErrors
from .qt import QTimer, QObject, Qt, QMenu
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


COMPLEXITY_PIXMAPS = {
    'A': 'complexity-a.png',
    'B': 'complexity-b.png',
    'C': 'complexity-c.png',
    'D': 'complexity-d.png',
    'E': 'complexity-e.png',
    'F': 'complexity-f.png'}



class PyflakesAttributes:

    """Holds all the attributes associated with pyflakes results"""

    def __init__(self):
        self.messages = []      # Complains
        self.ccMessages = []
        self.changed = False

    def hasComplains(self):
        """True if there are any complains"""
        if self.messages:
            return True
        return False


class PyflakesViewer(QObject):

    """The pyflakes viewer"""

    def __init__(self, editorsManager, uiLabel, ccLabel, parent=None):
        QObject.__init__(self, parent)

        self.__editorsManager = editorsManager
        self.__uiLabel = uiLabel
        self.__ccLabel = ccLabel
        self.setAnalysisNotAvailable(self.__uiLabel, self.__ccLabel)

        self.__editorsManager.currentChanged.connect(self.__onTabChanged)
        self.__editorsManager.sigTabClosed.connect(self.__onTabClosed)
        self.__editorsManager.sigBufferSavedAs.connect(self.__onSavedBufferAs)
        self.__editorsManager.sigFileTypeChanged.connect(
            self.__onFileTypeChanged)

        self.__flakesResults = {}  # UUID -> PyflakesAttributes
        self.__currentUUID = None
        self.__updateTimer = QTimer(self)
        self.__updateTimer.setSingleShot(True)
        self.__updateTimer.timeout.connect(self.__updateView)

        # Context menu for the messages icon
        self.__uiLabel.setContextMenuPolicy(Qt.CustomContextMenu)
        self.__uiLabel.customContextMenuRequested.connect(
            self.__showPyflakesContextMenu)
        self.__uiLabel.doubleClicked.connect(self.__jumpToFirstMessage)

        # Context menu for the CC icon
        self.__ccLabel.setContextMenuPolicy(Qt.CustomContextMenu)
        self.__ccLabel.customContextMenuRequested.connect(
            self.__showCCContextMenu)
        self.__ccLabel.doubleClicked.connect(self.__jumpToFirstCC)

    def __onTabChanged(self, index):
        """Triggered when another tab becomes active"""
        # If the timer is still active that means the tab was switched before
        # the handler had a chance to work. Therefore update the previous tab
        # first if so.
        if self.__updateTimer.isActive():
            self.__updateTimer.stop()
            self.__updateView()

        # Now, switch the pyflakes browser to the new tab
        if index == -1:
            widget = self.__editorsManager.currentWidget()
        else:
            widget = self.__editorsManager.getWidgetByIndex(index)
        if widget is None:
            self.__currentUUID = None
            return

        if widget.getType() not in [MainWindowTabWidgetBase.PlainTextEditor,
                                    MainWindowTabWidgetBase.VCSAnnotateViewer]:
            self.__currentUUID = None
            self.setAnalysisNotAvailable(self.__uiLabel, self.__ccLabel)
            return

        # This is text editor, detect the file type
        if not isPythonMime(widget.getMime()):
            self.__currentUUID = None
            self.setAnalysisNotAvailable(self.__uiLabel, self.__ccLabel)
            return

        # This is a python file, check if we already have the parsed info in
        # the cache
        uuid = widget.getUUID()
        self.__currentUUID = uuid
        if uuid in self.__flakesResults:
            # We have it, change the icon and the tooltip correspondingly
            results = self.__flakesResults[uuid].messages
            ccResults = self.__flakesResults[uuid].ccMessages
            self.setAnalysisResults(self.__uiLabel, results,
                                    self.__ccLabel, ccResults, None)
            return

        # It is first time we are here, create a new
        editor = widget.getEditor()
        editor.textChanged.connect(self.__onBufferChanged)
        editor.cursorPositionChanged.connect(self.__cursorPositionChanged)

        results, ccResults = getBufferErrors(editor.text)
        attributes = PyflakesAttributes()
        attributes.messages = results
        attributes.ccMessages = ccResults
        attributes.changed = False
        self.__flakesResults[uuid] = attributes
        self.__currentUUID = uuid

        self.setAnalysisResults(self.__uiLabel, results,
                                self.__ccLabel, ccResults, editor)

    def __cursorPositionChanged(self):
        """Triggered when a cursor position is changed"""
        if self.__updateTimer.isActive():
            # If a file is very large and the cursor is moved
            # straight after changes this will delay the update till
            # the real pause.
            self.__updateTimer.stop()
            self.__updateTimer.start(1500)

    def __onBufferChanged(self):
        """Triggered when a change in the buffer is identified"""
        if self.__currentUUID is None:
            return
        widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
        if widget is None:
            return
        if self.parent().debugMode:
            return

        self.__updateTimer.stop()
        if self.__currentUUID in self.__flakesResults:
            if not self.__flakesResults[self.__currentUUID].changed:
                self.__flakesResults[self.__currentUUID].changed = True
                self.setAnalysisWaiting(self.__uiLabel, self.__ccLabel)
        self.__updateTimer.start(1500)

    def __updateView(self):
        """Updates the view when a file is changed"""
        self.__updateTimer.stop()
        if self.__currentUUID is None:
            return
        widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
        if widget is None:
            return

        if not self.__flakesResults[self.__currentUUID].changed:
            return

        editor = widget.getEditor()
        results, ccResults = getBufferErrors(editor.text)

        self.__flakesResults[self.__currentUUID].messages = results
        self.__flakesResults[self.__currentUUID].ccMessages = ccResults
        self.__flakesResults[self.__currentUUID].changed = False

        self.setAnalysisResults(self.__uiLabel, results,
                                self.__ccLabel, ccResults, editor)

    def __onTabClosed(self, uuid):
        """Triggered when a tab is closed"""
        if uuid in self.__flakesResults:
            self.setAnalysisNotAvailable(self.__uiLabel, self.__ccLabel)
            del self.__flakesResults[uuid]

    def __onSavedBufferAs(self, fileName, uuid):
        """Triggered when a file is saved with a new name"""
        if uuid in self.__flakesResults:
            if not isPythonFile(fileName):
                # It's not a python file anymore
                self.__currentUUID = None
                del self.__flakesResults[uuid]
                self.setAnalysisNotAvailable(self.__uiLabel, self.__ccLabel)

    def __onFileTypeChanged(self, fileName, uuid, newFileType):
        """Triggered when the current buffer file type is changed, e.g. .cgi"""
        del fileName    # unused argument
        if isPythonMime(newFileType):
            # The file became a python one
            if uuid not in self.__flakesResults:
                self.__onTabChanged(-1)
        else:
            if uuid in self.__flakesResults:
                # It's not a python file any more
                if uuid == self.__currentUUID:
                    self.__currentUUID = None

                del self.__flakesResults[uuid]
                self.setAnalysisNotAvailable(self.__uiLabel, self.__ccLabel)

    def __showPyflakesContextMenu(self, pos):
        """Triggered when the icon context menu is requested"""
        if self.__currentUUID is None:
            return
        if self.__currentUUID not in self.__flakesResults:
            return

        messages = self.__flakesResults[self.__currentUUID].messages
        if not messages:
            return

        # Check that there is at least one non -1 lineno message
        lineNumbers = list(messages.keys())
        for lineno in lineNumbers:
            if lineno > 0:
                break
        else:
            return

        # OK, we have something to show
        lineNumbers.sort()
        contextMenu = QMenu(self.__uiLabel)
        for lineno in lineNumbers:
            if lineno > 0:
                for item in messages[lineno]:
                    act = contextMenu.addAction(
                        getIcon('pyflakesmsgmarker.png'),
                        "Line " + str(lineno) + ": " + item)
                    act.setData(lineno)
        contextMenu.triggered.connect(self.__onContextMenu)
        contextMenu.popup(self.__uiLabel.mapToGlobal(pos))

    def __showCCContextMenu(self, pos):
        """Triggered when the cc icon context menu is requested"""
        if self.__currentUUID is None:
            return
        if self.__currentUUID not in self.__flakesResults:
            return

        count = 0
        contextMenu = QMenu(self.__ccLabel)
        for item in self.__flakesResults[self.__currentUUID].ccMessages:
            complexity = cc_rank(item.complexity)

            if complexity != 'A':
                count += 1
                title = complexity + '(' + str(item.complexity) + ') ' + \
                        item.fullname
                if item.letter in ('F', 'M'):
                    title += '()'
                act = contextMenu.addAction(getIcon('ccmarker.png'), title)
                act.setData(item.lineno)
        if count > 0:
            contextMenu.triggered.connect(self.__onContextMenu)
            contextMenu.popup(self.__ccLabel.mapToGlobal(pos))
        else:
            del contextMenu

    def __jumpToFirstCC(self):
        """Double click on the cc icon"""
        if self.__currentUUID is None:
            return
        if self.__currentUUID not in self.__flakesResults:
            return

        if self.__flakesResults[self.__currentUUID].ccMessages:
            firstItem = self.__flakesResults[self.__currentUUID].ccMessages[0]
            widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
            if widget:
                self.__editorsManager.jumpToLine(firstItem.lineno)

    def __onContextMenu(self, act):
        """Triggered when a context menu item is selected"""
        if self.__currentUUID is None:
            return
        widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
        if widget:
            self.__editorsManager.jumpToLine(act.data())

    def __jumpToFirstMessage(self):
        """Double click on the icon"""
        if self.__currentUUID is None:
            return
        if self.__currentUUID not in self.__flakesResults:
            return

        messages = self.__flakesResults[self.__currentUUID].messages
        if not messages:
            return

        widget = self.__editorsManager.getWidgetByUUID(self.__currentUUID)
        if widget:
            lineNumbers = list(messages.keys())
            lineNumbers.sort()
            for lineno in lineNumbers:
                if lineno > 0:
                    self.__editorsManager.jumpToLine(lineno)
                    break

    @staticmethod
    def setAnalysisResults(label, results, ccLabel, ccResults, editor):
        """Displays the appropriate icon:

        - pyflakes has no complains
        - pyflakes found errors
        """
        if editor is not None:
            editor.clearAnalysisMessages()
            editor.setAnalysisMessages(results, ccResults)

        if results:
            # There are complains
            complains = "Buffer checked: there are pyflakes complains<br/>"
            lineNumbers = list(results.keys())
            lineNumbers.sort()
            for lineNo in lineNumbers:
                for item in results[lineNo]:
                    complains += '<br/>'
                    if lineNo == -1:
                        # Special case: compilation error
                        complains += escape(item)
                    else:
                        complains += "Line " + str(lineNo) + \
                                     ": " + escape(item)
            label.setToolTip(complains.replace(' ', '&nbsp;'))
            label.setPixmap(getPixmap('flakeserrors.png'))
        else:
            # There are no complains
            label.setToolTip('Buffer checked: no pyflakes complains')
            label.setPixmap(getPixmap('flakesok.png'))

        if ccResults:
            complains = 'Buffer cyclomatic complexity:<br/>'
            worstComplexity = 'A'
            for item in ccResults:
                complexity = cc_rank(item.complexity)
                worstComplexity = max(complexity, worstComplexity)

                if complexity != 'A':
                    complains += '<br/>' + complexity + \
                                 '(' + str(item.complexity) + ') ' + \
                                 escape(item.fullname)
                    if item.letter in ('F', 'M'):
                        complains += '()'

            if worstComplexity == 'A':
                ccLabel.setToolTip('Buffer cyclomatic complexity: no complains')
            else:
                ccLabel.setToolTip(complains.replace(' ', '&nbsp;'))
            ccLabel.setPixmap(getPixmap(COMPLEXITY_PIXMAPS[worstComplexity]))
        else:
            ccLabel.setToolTip('No complexity information available')
            ccLabel.setPixmap(getPixmap('ccmarker.png'))

    @staticmethod
    def setAnalysisWaiting(label, ccLabel):
        """Displays the waiting for a time slice to start checking icon"""
        label.setToolTip('File is modified: '
                         'pyflakes is waiting for time slice')
        label.setPixmap(getPixmap('flakesmodified.png'))

        ccLabel.setToolTip('File is modified: '
                           'radon is waiting for time slice')
        ccLabel.setPixmap(getPixmap('flakesmodified.png'))

    @staticmethod
    def setAnalysisNotAvailable(label, ccLabel):
        """Displays the appropriate icon that pyflakes is not available"""
        label.setToolTip('Not a python file: pyflakes is sleeping')
        label.setPixmap(getPixmap('flakessleep.png'))

        ccLabel.setToolTip('Not a python file: radon is sleeping')
        ccLabel.setPixmap(getPixmap('flakessleep.png'))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Project browser model"""

from utils.globals import GlobalData
from utils.project import CodimensionProject
from utils.settings import Settings
from .qt import Qt
from .viewitems import TreeViewDirectoryItem
from .browsermodelbase import BrowserModelBase


class ProjectBrowserModel(BrowserModelBase):

    """Class implementing the project browser model"""

    def __init__(self, parent):
        self.__mainWindow = parent
        BrowserModelBase.__init__(self, "Name", self.__mainWindow)
        self.setTooltips(Settings()['projectTooltips'])
        self.populateModel()
        GlobalData().project.sigProjectChanged.connect(self.__onProjectChanged)

    def populateModel(self):
        """Populates the project browser model"""
        self.clear()
        project = self.globalData.project
        if project.isLoaded():
            projectDir = project.getProjectDir()
            newItem = TreeViewDirectoryItem(self.rootItem, projectDir)
            newItem.needVCSStatus = True
            self.addItem(newItem)

    def __onProjectChanged(self, what):
        """Triggered when a project is changed"""
        if what == CodimensionProject.CompleteProject:
            self.populateModel()

    def __tooltipRoleData(self, item, indicator):
        """Provides the data for the tooltip role"""
        docstringPart = None
        if self.showTooltips and item.toolTip != "":
            docstringPart = item.toolTip

        vcsPart = None
        if indicator:
            if item.vcsStatus.message:
                vcsPart = item.vcsStatus.message
            else:
                vcsPart = indicator.defaultTooltip

        if docstringPart and vcsPart:
            return docstringPart + "\n\nVCS status: " + vcsPart
        if docstringPart:
            return docstringPart
        if vcsPart:
            return "VCS status: " + vcsPart
        return None

    def data(self, index, role):
        """Extention to modify the background and tooltips for the browser"""
        if not index.isValid():
            return None

        item = index.internalPointer()
        if item.vcsStatus:
            indicator = self.__mainWindow.vcsManager.getStatusIndicator(
                item.vcsStatus)
            if role == Qt.TextColorRole:
                if indicator and indicator.foregroundColor:
                    return indicator.foregroundColor
            elif role == Qt.BackgroundColorRole:
                if indicator and indicator.backgroundColor:
                    return indicator.backgroundColor
            elif role == Qt.ToolTipRole:
                return self.__tooltipRoleData(item, indicator)

        return BrowserModelBase.data(self, index, role)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2019  Sergey V. Satskiy sergey.satskiy@tuta.io
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Rendered plantuml diagram cache"""

# pylint: disable=C0305
# pylint: disable=W0702
# pylint: disable=W0703

import sys
import subprocess
import os.path
import logging
import datetime
import hashlib
from distutils.spawn import find_executable
from ui.qt import QThread, pyqtSignal, QObject
from .fileutils import loadJSON, saveJSON, saveToFile

CACHE_FILE_NAME = 'cachemap.json'


def getPlantUMLJarPath():
    """Provides the full path to the plantUML jar file"""
    if find_executable('java') is None:
        return None
    exeDir = os.path.dirname(os.path.realpath(sys.argv[0]))
    plantUMLPath = os.path.sep.join([os.path.dirname(exeDir), 'plantuml'])
    for item in os.listdir(plantUMLPath):
        if item.startswith('plantuml.') and item.endswith('.jar'):
            return plantUMLPath + os.path.sep + item
    return None
JAR_PATH = getPlantUMLJarPath()


def __removeEmptyForward(items, startIndex, length):
    """Removes empty lines at the beginning"""
    toRemove = []
    for index in range(startIndex, length):
        if not items[index].strip():
            toRemove.append(index)
    for index in range(len(toRemove)):
        items.pop(toRemove[-(index + 1)])
    return items


def __removeEmptyBackward(items, endIndex):
    """Removes empty lines at the end"""
    while not items[endIndex].strip():
        items.pop(endIndex)
        endIndex -= 1
    return items


def normalizePlantumlSource(source):
    """Normalizes the diagram source"""
    lines = source.strip().splitlines()
    length = len(lines)
    if length == 1:
        if lines[0].startswith('@start'):
            # end is missed
            return '\n'.join([lines[0], '@end' + lines[0][6:]])
        if lines[0].startswith('@end'):
            # start is missed
            return'\n'.join(['@start' + lines[0][4:], lines[0]])
        # missed both
        return '\n'.join(['@startuml', lines[0], '@enduml'])

    # More than one line
    startFound = lines[0].startswith('@start')
    endFound = lines[-1].lstrip().startswith('@end')

    if startFound and endFound:
        # both found
        # remove empty lines before and after
        __removeEmptyForward(lines, 1, length)
        __removeEmptyBackward(lines, len(lines) - 2)
        lines[0] = lines[0].strip()
        lines[-1] = lines[-1].strip()
    elif startFound:
        # start is here, end is missed
        __removeEmptyForward(lines, 1, length)
        lines[0] = lines[0].strip()
        lines.append('@end' + lines[0][6:])
    elif endFound:
        # start is missed, end is here
        __removeEmptyBackward(lines, length - 2)
        lines[-1] = lines[-1].strip()
        lines.insert(0, '@start' + lines[-1][4:])
    else:
        # both missed
        __removeEmptyForward(lines, 0, length)
        __removeEmptyBackward(lines, len(lines) - 1)
        lines.insert(0, '@startuml')
        lines.append('@enduml')

    return '\n'.join(lines)


class PlantUMLRenderer(QThread):

    """Runs plantuml"""

    sigFinishedOK = pyqtSignal(str, str, str)       # md5, uuid, file
    sigFinishedError = pyqtSignal(str, str)         # md5, file

    def __init__(self, parent=None):
        QThread.__init__(self, parent)
        self.__source = None
        self.__md5 = None
        self.__uuid = None
        self.__fName = None

    def get(self, source, md5, fName, uuid):
        """Initiates rendering the diagram"""
        self.__source = source
        self.__md5 = md5
        self.__uuid = uuid
        self.__fName = fName
        self.start()

    @staticmethod
    def safeUnlink(fName):
        """Safe file removal"""
        try:
            os.unlink(fName)
        except:
            pass

    def run(self):
        """Runs plantUML"""
        srcFile = self.__fName[:-3] + 'txt'
        try:
            # Run plantUML
            saveToFile(srcFile, self.__source)
            retCode = subprocess.call(['java', '-jar', JAR_PATH,
                                       '-charset', 'utf-8', '-nometadata',
                                       srcFile],
                                      stdout=subprocess.DEVNULL,
                                      stderr=subprocess.DEVNULL)
            self.safeUnlink(srcFile)

            if retCode == 0:
                self.sigFinishedOK.emit(self.__md5, self.__uuid, self.__fName)
            else:
                self.sigFinishedError.emit(self.__md5, self.__fName)
        except Exception as exc:
            logging.error('Cannot render a plantUML diagram: %s', str(exc))
            self.safeUnlink(srcFile)
            self.sigFinishedError.emit(self.__md5, self.__fName)



class PlantUMLCache(QObject):

    """The plantUML render cache"""

    sigRenderReady = pyqtSignal(str, str)    # uuid, file

    def __init__(self, cacheDir):
        QObject.__init__(self)

        self.__md5ToFileName = {}
        self.__threads = {}
        self.__cacheDir = os.path.normpath(cacheDir) + os.path.sep

        if os.path.exists(self.__cacheDir):
            if os.path.isdir(self.__cacheDir):
                if os.access(self.__cacheDir, os.W_OK):
                    self.__loadCache()
                    self.__saveCache()
                else:
                    logging.error('The plantUML render cache directory (%s) '
                                  'does not have write permissions. There will '
                                  'be no plantUML rendering', self.__cacheDir)
                    self.__cacheDir = None
            else:
                logging.error('The plantUML render cache directory path (%s) '
                              'exists and is not a directory. There will be no '
                              'pluntUML rendering', self.__cacheDir)
                self.__cacheDir = None
        else:
            # Try to create the dir
            try:
                os.mkdir(self.__cacheDir)
            except Exception as exc:
                logging.error('Error creating pluntUML render cache directory '
                              '%s: %s', self.__cacheDir, str(exc))
                logging.error('There will be no plantUML rendering')
                self.__cacheDir = None

    def __loadCache(self):
        """Loads the cache from the disk files"""
        if self.__cacheDir is None:
            return

        # Remove too old files
        now = datetime.datetime.now()
        limit = now - datetime.timedelta(days=1)
        limit = limit.timestamp()
        for item in os.listdir(self.__cacheDir):
            if item == CACHE_FILE_NAME:
                continue
            if os.path.isfile(self.__cacheDir + item):
                modtime = os.path.getmtime(self.__cacheDir + item)
                if modtime < limit:
                    try:
                        os.unlink(self.__cacheDir + item)
                    except Exception as exc:
                        logging.error('Error removing obsolete plantUML '
                                      'render file (%s): %s',
                                      self.__cacheDir + item, str(exc))

        if os.path.exists(self.__cacheDir + CACHE_FILE_NAME):
            prevCache = loadJSON(self.__cacheDir + CACHE_FILE_NAME,
                                 'plantUML render cache map', None)
            for item in prevCache.items():
                if os.path.exists(item[1]):
                    self.__md5ToFileName[item[0]] = item[1]

    def __saveCache(self):
        """Saves the cache to the disk"""
        if self.__cacheDir is None:
            return
        dictToSave = {}
        for item in self.__md5ToFileName.items():
            if item[1] is not None:
                dictToSave[item[0]] = item[1]
        saveJSON(self.__cacheDir + CACHE_FILE_NAME, dictToSave,
                 'plantUML render cache map')

    def onRenderOK(self, md5, uuid, fName):
        """Render saved successfully"""
        self.__md5ToFileName[md5] = fName
        self.__onThreadFinish(fName)
        self.__saveCache()
        self.sigRenderReady.emit(uuid, fName)

    def onRenderError(self, md5, fName):
        """Error rendering"""
        self.__md5ToFileName[md5] = None
        self.__onThreadFinish(fName)

    def __onThreadFinish(self, fName):
        """Cleans up after a retrieval thread"""
        thread = self.__threads.get(fName, None)
        if thread is not None:
            self.__disconnectThread(thread)
            thread.wait()
            self.__threads.pop(fName)

    def __connectThread(self, thread):
        """Connects the thread signals"""
        thread.sigFinishedOK.connect(self.onRenderOK)
        thread.sigFinishedError.connect(self.onRenderError)

    def __disconnectThread(self, thread):
        """Connects the thread signals"""
        try:
            thread.sigFinishedOK.disconnect(self.onRenderOK)
            thread.sigFinishedError.disconnect(self.onRenderError)
        except:
            pass

    def getResource(self, source, uuid):
        """Provides the rendered file name

        If None => no rendering will be done
        Otherwise the ready-to-use file or where the pic is expected
        """
        if self.__cacheDir is None or JAR_PATH is None:
            return None

        normSource = normalizePlantumlSource(source)
        md5 = hashlib.md5(normSource.encode('utf-8')).hexdigest()
        if md5 in self.__md5ToFileName:
            return self.__md5ToFileName[md5]

        basename = md5 + '.png'
        fName = self.__cacheDir + basename
        if fName in self.__threads:
            # Reject double request
            return fName

        thread = PlantUMLRenderer()
        self.__threads[fName] = thread
        self.__connectThread(thread)
        thread.get(normSource, md5, fName, uuid)
        return fName



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Python code parser errors dialog"""


from os.path import exists, basename
from utils.globals import GlobalData
from utils.fileutils import isPythonFile
from utils.colorfont import getZoomedMonoFont
from .qt import (Qt, QDialog, QTextEdit, QDialogButtonBox, QVBoxLayout,
                 QSizePolicy)
from .labels import FitLabel


class ParserErrorsDialog(QDialog):

    """Python code parser errors dialog implementation"""

    def __init__(self, fileName, info=None, parent=None):
        QDialog.__init__(self, parent)

        if info is None:
            if not exists(fileName):
                raise Exception('Cannot open ' + fileName)

            if not isPythonFile(fileName):
                raise Exception('Unexpected file type (' + fileName +
                                '). A python file is expected.')

        self.__createLayout(fileName, info)
        self.setWindowTitle('Lexer/parser errors: ' + basename(fileName))
        self.show()

    def __createLayout(self, fileName, info):
        """Creates the dialog layout"""
        self.resize(600, 220)
        self.setSizeGripEnabled(True)

        verticalLayout = QVBoxLayout(self)

        # Info label
        infoLabel = FitLabel(self)
        sizePolicy = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            infoLabel.sizePolicy().hasHeightForWidth())
        infoLabel.setSizePolicy(sizePolicy)
        infoLabel.setText('Lexer/parser errors for ' + fileName)
        verticalLayout.addWidget(infoLabel)

        # Result window
        resultEdit = QTextEdit(self)
        resultEdit.setTabChangesFocus(False)
        resultEdit.setAcceptRichText(False)
        resultEdit.setReadOnly(True)
        resultEdit.setFont(getZoomedMonoFont())
        if info is not None:
            modInfo = info
        else:
            modInfo = GlobalData().briefModinfoCache.get(fileName)
        if modInfo.isOK:
            resultEdit.setText('No errors found')
        else:
            resultEdit.setText('\n'.join(modInfo.lexerErrors +
                                         modInfo.errors))
        verticalLayout.addWidget(resultEdit)

        # Buttons
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Close)
        verticalLayout.addWidget(buttonBox)

        buttonBox.rejected.connect(self.close)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Dialog to show a single variable"""


from ui.qt import (Qt, QDialog, QDialogButtonBox, QVBoxLayout, QLabel,
                   QGridLayout, QTextEdit)
from utils.pixmapcache import getIcon
from ui.labels import HeaderLabel
from utils.globals import GlobalData
from utils.colorfont import getZoomedMonoFont


class ViewVariableDialog(QDialog):

    """Dialog all the properties of a variable"""

    def __init__(self, varName,
                 varType, varValue, isGlobal, parent=None):
        QDialog.__init__(self, parent)

        if varName.endswith("."):
            varName = varName[:-1]

        if isGlobal:
            self.setWindowTitle("Global variable '" + varName + "'")
            self.setWindowIcon(getIcon("globvar.png"))
        else:
            self.setWindowTitle("Local variable '" + varName + "'")
            self.setWindowIcon(getIcon( "locvar.png"))
        self.__createLayout(varName, varType, varValue, isGlobal)

    def __createLayout(self, varName, varType, varValue, isGlobal):
        """Creates the dialog layout"""
        varTypeParts = varType.split()
        if varTypeParts[0].lower() in ["string", "unicode", "qstring"]:
            length = str(len(varValue))
            lines = str(len(varValue.splitlines()))
            varType = varType.split("(")[0].strip() + \
                      " (lines: " + lines + ", characters: " + length + ")"

        self.resize(600, 250)
        self.setSizeGripEnabled(True)

        # Top level layout
        layout = QVBoxLayout(self)

        gridLayout = QGridLayout()
        gridLayout.setSpacing(4)
        varScopeLabel = QLabel("Scope:", self)
        gridLayout.addWidget(varScopeLabel, 0, 0, Qt.AlignCenter)
        varScopeValue = HeaderLabel('Global' if isGlobal else 'Local',
                                    parent=self)
        varScopeValue.setToolTip("Double click to copy")
        font = varScopeValue.font()
        font.setFamily(GlobalData().skin['monoFont'].family())
        gridLayout.addWidget(varScopeValue, 0, 1)

        varNameLabel = QLabel("Name:", self)
        gridLayout.addWidget(varNameLabel, 1, 0, Qt.AlignCenter)
        varNameValue = HeaderLabel(varName, parent=self)
        varNameValue.setToolTip("Double click to copy")
        gridLayout.addWidget(varNameValue, 1, 1)

        varTypeLabel = QLabel("Type:", self)
        gridLayout.addWidget(varTypeLabel, 2, 0, Qt.AlignCenter)
        varTypeValue = HeaderLabel(varType, parent=self)
        varTypeValue.setToolTip("Double click to copy")
        gridLayout.addWidget(varTypeValue, 2, 1)

        varValueLabel = QLabel("Value:", self)
        gridLayout.addWidget(varValueLabel, 3, 0, Qt.AlignTop)
        varValueValue = QTextEdit()
        varValueValue.setReadOnly(True)
        varValueValue.setFont(getZoomedMonoFont())
        # varValueValue.setLineWrapMode(QTextEdit.NoWrap)
        varValueValue.setAcceptRichText(False)
        varValueValue.setPlainText(varValue)
        gridLayout.addWidget(varValueValue, 3, 1)
        layout.addLayout(gridLayout)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Ok)
        self.__OKButton = buttonBox.button(QDialogButtonBox.Ok)
        self.__OKButton.setDefault(True)
        buttonBox.accepted.connect(self.close)
        buttonBox.rejected.connect(self.close)
        layout.addWidget(buttonBox)

        varValueValue.setFocus()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Redirectors for stdout/stderr streams"""

# pylint: disable=no-self-use, unused-argument

from protocol_cdm_dbg import METHOD_STDOUT, METHOD_STDERR
from cdm_dbg_utils import sendJSONCommand


class OutStreamRedirector():

    """Wraps a socket object with a file interface"""

    def __init__(self, sock, isStdout, procid):
        self.closed = False
        self.sock = sock
        self.isStdout = isStdout
        self.procid = procid

    def close(self, closeit=False):
        """Closes the file. closeit != 0 => debugger requested it"""
        if closeit and not self.closed:
            self.sock.close()
            self.closed = True

    def flush(self):
        """Does nothing because there is no buffering"""
        pass

    def isatty(self):
        """Indicates whether a tty interface is supported"""
        return False

    def fileno(self):
        """Provides the file number"""
        try:
            return self.sock.socketDescriptor()
        except Exception as exc:
            return -1

    def read_p(self, size=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def read(self, size=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def readline_p(self, size=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def readlines(self, sizehint=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def readline(self, sizehint=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def seekable(self):
        """Checks if the stream is seekable"""
        return False

    def seek(self, offset, whence=0):
        """Seek is not supported"""
        raise IOError((29, '[Errno 29] Illegal seek'))

    def tell(self):
        """Tell is not supported"""
        raise IOError((29, '[Errno 29] Illegal seek'))

    def truncate(self, size=-1):
        """Truncates is not supported"""
        raise IOError((29, '[Errno 29] Illegal seek'))

    def write(self, data):
        """Writes a string to the file"""
        method = METHOD_STDERR
        if self.isStdout:
            method = METHOD_STDOUT
        sendJSONCommand(self.sock, method, self.procid, {'text': data})

    def writelines(self, lines):
        """Writes a list of strings to the file"""
        self.write(''.join(lines))


# The OutStreamCollector is used in case of the Exec() request.
# The Exec() supposes that there could be both stdout and stderr
# printouts. They need to be collected as a combined output and sent back to
# the IDE. So an instance of the OutStreamCollector is used to temporary
# substitute the sys.stdout and sys.stderr.
class OutStreamCollector():

    """Collects output with a file interface"""

    def __init__(self):
        self.buf = ""

    def close(self):
        """Closes the file"""
        pass

    def flush(self):
        """Does nothing because there is no buffering"""
        pass

    def isatty(self):
        """Indicates whether a tty interface is supported"""
        return False

    def fileno(self):
        """Provides the file number"""
        return -1

    def readable(self):
        """Checks if the stream is readable"""
        return False

    def read_p(self, size=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def read(self, size=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def readline_p(self, size=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def readlines(self, sizehint=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def readline(self, sizehint=-1):
        """Read is not supported"""
        raise IOError((9, '[Errno 9] Bad file descriptor'))

    def seekable(self):
        """Checks if the stream is seekable"""
        return False

    def seek(self, offset, whence=0):
        """Seek is not supported"""
        raise IOError((29, '[Errno 29] Illegal seek'))

    def tell(self):
        """Tell is not supported"""
        raise IOError((29, '[Errno 29] Illegal seek'))

    def truncate(self, size=-1):
        """Truncates is not supported"""
        raise IOError((29, '[Errno 29] Illegal seek'))

    def writable(self):
        """Check if a stream is writable"""
        return True

    def write(self, data):
        """Writes a string to the file"""
        self.buf += data

    def writelines(self, lines):
        """Writes a list of strings to the file"""
        self.write(''.join(lines))



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Dialog with a list of modified but unsaved files"""

from ui.qt import (Qt, QDialog, QDialogButtonBox, QVBoxLayout, QSizePolicy,
                   QLabel, QHBoxLayout, QTreeWidget, QAbstractItemView,
                   QTreeWidgetItem, QWidget)
from ui.itemdelegates import NoOutlineHeightDelegate
from utils.pixmapcache import getIcon, getPixmap
from utils.fileutils import getFileProperties, isPythonMime
from utils.globals import GlobalData


class ModifiedUnsavedDialog(QDialog):

    """Dialog with a list of modified but unsaved files implementation"""

    # See utils.run for runParameters
    def __init__(self, files, action, parent=None):
        QDialog.__init__(self, parent)

        count = len(files)
        if count >= 2:
            title = str(count) + " project files are modified and not saved"
        else:
            title = "1 project file modified and not saved"
        self.setWindowTitle(title)
        self.setWindowIcon(getIcon('warning.png'))
        self.__createLayout(action, title, files)

    def __createLayout(self, action, title, files):
        """Creates the dialog layout"""
        self.resize(400, 300)
        self.setSizeGripEnabled(True)

        # Top level layout
        layout = QVBoxLayout(self)

        # Pixmap and the message
        topLayout = QHBoxLayout()
        pixmap = QLabel()
        pixmap.setPixmap(getPixmap('warning.png'))
        topLayout.addWidget(pixmap)
        hSpacer = QWidget()
        hSpacer.setFixedSize(15, 15)
        topLayout.addWidget(hSpacer)
        message = QLabel("All the project files must be "
                         "saved before start debugging")
        message.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        message.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        message.setWordWrap(True)
        topLayout.addWidget(message)
        layout.addLayout(topLayout)

        vSpacer = QWidget()
        vSpacer.setFixedSize(15, 15)
        layout.addWidget(vSpacer)

        layout.addWidget(QLabel(title + ":"))
        filesList = QTreeWidget()
        filesList.setRootIsDecorated(False)
        filesList.setAlternatingRowColors(True)
        filesList.setUniformRowHeights(True)
        filesList.setItemsExpandable(False)
        filesList.setItemDelegate(NoOutlineHeightDelegate(4))
        filesList.setSelectionMode(QAbstractItemView.NoSelection)
        filesList.setHeaderHidden(True)
        for item in files:
            fileName = item[0]
            fileItem = QTreeWidgetItem([fileName])
            fileType, icon, _ = getFileProperties(fileName)
            fileItem.setIcon(0, icon)
            if isPythonMime(fileType):
                info = GlobalData().briefModinfoCache.get(fileName)
                fileItem.setToolTip(0, info.docstring.text if info.docstring
                                    else '')
            filesList.addTopLevelItem(fileItem)
        layout.addWidget(filesList)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox()
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Cancel)
        continueButton = buttonBox.addButton(action,
                                             QDialogButtonBox.ActionRole)
        continueButton.setDefault(True)
        layout.addWidget(buttonBox)

        continueButton.clicked.connect(self.accept)
        buttonBox.rejected.connect(self.close)
        continueButton.setFocus()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Common functionality of various browser models"""

import sys
import os
import logging
from utils.fileutils import isPythonFile, isPythonMime
from utils.globals import GlobalData
from utils.pixmapcache import getIcon
from .qt import Qt, QAbstractItemModel, QModelIndex, QApplication, QCursor
from .viewitems import (TreeViewItem, TreeViewDirectoryItem, TreeViewFileItem,
                        TreeViewGlobalsItem, TreeViewImportsItem,
                        TreeViewFunctionsItem, TreeViewClassesItem,
                        TreeViewStaticAttributesItem, GlobalsItemType,
                        TreeViewInstanceAttributesItem, FileItemType,
                        TreeViewCodingItem, TreeViewImportItem,
                        TreeViewFunctionItem, TreeViewClassItem,
                        TreeViewDecoratorItem, TreeViewAttributeItem,
                        TreeViewGlobalItem, TreeViewWhatItem,
                        DirectoryItemType, SysPathItemType,
                        ImportsItemType, FunctionsItemType,
                        ClassesItemType, StaticAttributesItemType,
                        InstanceAttributesItemType, DecoratorItemType,
                        FunctionItemType, ClassItemType, ImportItemType)


class BrowserModelBase(QAbstractItemModel):

    """Class implementing the file system browser model"""

    def __init__(self, headerData, parent=None):
        QAbstractItemModel.__init__(self, parent)

        self.rootItem = TreeViewItem(None, headerData)
        self.globalData = GlobalData()
        self.projectTopLevelDirs = []
        self.showTooltips = True

    def setTooltips(self, switchOn):
        """Sets the tooltip mode: to show or not to show them"""
        self.showTooltips = switchOn

    def columnCount(self, parent=QModelIndex()):
        """Provides the number of columns"""
        if parent.isValid():
            return parent.internalPointer().columnCount()
        return self.rootItem.columnCount()

    def updateRootData(self, column, value):
        """Updates the root entry, i.e. header"""
        self.rootItem.setData(column, value)
        self.headerDataChanged.emit(Qt.Horizontal, column, column)

    def data(self, index, role):
        """Provides data of an item"""
        if not index.isValid():
            return None

        column = index.column()
        if role == Qt.DisplayRole:
            item = index.internalPointer()
            if column < item.columnCount():
                return item.data(column)
            if column == item.columnCount() and \
               column < self.columnCount(self.parent(index)):
                # This is for the case when an item under a multi-column
                # parent doesn't have a value for all the columns
                return ""
        elif role == Qt.DecorationRole:
            if column == 0:
                return index.internalPointer().getIcon()
        elif role == Qt.ToolTipRole:
            item = index.internalPointer()
            if column == 1 and item.path is not None:
                return item.path
            if self.showTooltips and column == 0 and item.toolTip != "":
                return item.toolTip
        return None

    def flags(self, index):
        """Provides the item flags"""
        if not index.isValid():
            return Qt.ItemIsEnabled
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    def headerData(self, section, orientation, role=Qt.DisplayRole):
        """Provides the header data"""
        if orientation == Qt.Horizontal and role == Qt.DisplayRole:
            if section >= self.rootItem.columnCount():
                return ""
            return self.rootItem.data(section)
        return None

    def index(self, row, column, parent=QModelIndex()):
        """Creates an index"""
        # The model/view framework considers negative values out-of-bounds,
        # however in python they work when indexing into lists. So make sure
        # we return an invalid index for out-of-bounds row/col
        if row < 0 or column < 0 or \
           row >= self.rowCount(parent) or \
           column >= self.columnCount(parent):
            return QModelIndex()

        if parent.isValid():
            parentItem = parent.internalPointer()
        else:
            parentItem = self.rootItem

        try:
            if not parentItem.populated:
                self.populateItem(parentItem)
            childItem = parentItem.child(row)
        except IndexError:
            return QModelIndex()

        if childItem:
            return self.createIndex(row, column, childItem)
        return QModelIndex()

    def buildIndex(self, rowPath):
        """Builds index for the path (path is like [1, 2, 1, 16])"""
        result = QModelIndex()
        for row in rowPath:
            result = self.index(row, 0, result)
        return result

    def parent(self, index):
        """Provides the index of the parent object"""
        if not index.isValid():
            return QModelIndex()

        childItem = index.internalPointer()
        parentItem = childItem.parent()

        if parentItem == self.rootItem:
            return QModelIndex()
        return self.createIndex(parentItem.row(), 0, parentItem)

    def totalRowCount(self):
        """Provides the total number of rows"""
        return self.rootItem.childCount()

    def rowCount(self, parent=QModelIndex()):
        """Provides the number of rows"""
        # Only the first column should have children
        if parent.column() > 0:
            return 0

        if not parent.isValid():
            return self.rootItem.childCount()

        parentItem = parent.internalPointer()
        if not parentItem.populated:    # lazy population
            self.populateItem(parentItem)
        return parentItem.childCount()

    def hasChildren(self, parent=QModelIndex()):
        """Returns True if the parent has children"""
        # Only the first column should have children
        if parent.column() > 0:
            return False
        if not parent.isValid():
            return self.rootItem.childCount() > 0

        if parent.internalPointer().lazyPopulation:
            return True
        return parent.internalPointer().childCount() > 0

    def clear(self):
        """Clears the model"""
        self.beginResetModel()
        self.rootItem.removeChildren()
        self.endResetModel()

    def item(self, index):
        """Provides a reference to an item"""
        if not index.isValid():
            return None
        return index.internalPointer()

    @staticmethod
    def _addItem(itm, parentItem):
        """Adds an item"""
        parentItem.appendChild(itm)

    def addItem(self, itm, parent=QModelIndex()):
        """Adds an item"""
        if not parent.isValid():
            parentItem = self.rootItem
        else:
            parentItem = parent.internalPointer()

        cnt = parentItem.childCount()
        self.beginInsertRows(parent, cnt, cnt)
        self._addItem(itm, parentItem)
        self.endInsertRows()

    def populateItem(self, parentItem, repopulate=False):
        """Populates an item's subtree"""
        if parentItem.itemType == DirectoryItemType:
            self.populateDirectoryItem(parentItem, repopulate)
        elif parentItem.itemType == SysPathItemType:
            self.populateSysPathItem(parentItem, repopulate)
        elif parentItem.itemType == FileItemType:
            self.populateFileItem(parentItem, repopulate)
        elif parentItem.itemType == GlobalsItemType:
            self.populateGlobalsItem(parentItem, repopulate)
        elif parentItem.itemType == ImportsItemType:
            self.populateImportsItem(parentItem, repopulate)
        elif parentItem.itemType == FunctionsItemType:
            self.populateFunctionsItem(parentItem, repopulate)
        elif parentItem.itemType == ClassesItemType:
            self.populateClassesItem(parentItem, repopulate)
        elif parentItem.itemType == ClassItemType:
            self.populateClassItem(parentItem, repopulate)
        elif parentItem.itemType == StaticAttributesItemType:
            self.populateStaticAttributesItem(parentItem, repopulate)
        elif parentItem.itemType == InstanceAttributesItemType:
            self.populateInstanceAttributesItem(parentItem, repopulate)
        elif parentItem.itemType == FunctionItemType:
            self.populateFunctionItem(parentItem, repopulate)
        elif parentItem.itemType == ImportItemType:
            self.populateImportItem(parentItem, repopulate)
        parentItem.populated = True

    def populateDirectoryItem(self, parentItem, repopulate=False):
        """Populates a directory item's subtree"""
        path = parentItem.getPath()
        if not os.path.exists(path):
            return

        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        try:
            items = os.listdir(path)
        except Exception as exc:
            QApplication.restoreOverrideCursor()
            logging.error("Cannot populate directory. " + str(exc))
            return

        excludes = ['.svn', '.cvs', '.hg', '.git']
        items = [itm for itm in items if itm not in excludes]
        if parentItem.needVCSStatus:
            # That's the project browser. Filter out what not needed.
            excludeFunctor = GlobalData().project.shouldExclude
            items = [itm for itm in items if not excludeFunctor(itm)]

        pathsToRequest = []
        if items:
            infoSrc = self.globalData.briefModinfoCache

            if repopulate:
                self.beginInsertRows(self.createIndex(parentItem.row(),
                                                      0, parentItem),
                                     0, len(items) - 1)
            path = os.path.realpath(path) + os.path.sep
            for item in items:
                fullPath = path + item
                if os.path.isdir(fullPath):
                    node = TreeViewDirectoryItem(parentItem, fullPath, False)
                    if parentItem.needVCSStatus:
                        pathsToRequest.append(fullPath + os.path.sep)
                else:
                    node = TreeViewFileItem(parentItem, fullPath)
                    if parentItem.needVCSStatus:
                        pathsToRequest.append(fullPath)
                    if isPythonMime(node.fileType):
                        modInfo = infoSrc.get(fullPath)
                        node.toolTip = ""
                        if modInfo.docstring is not None:
                            node.toolTip = modInfo.docstring.text

                        if modInfo.isOK == False:
                            # Substitute icon and change the tooltip
                            node.icon = getIcon('filepythonbroken.png')
                            if node.toolTip != "":
                                node.toolTip += "\n\n"
                            node.toolTip += "Parsing errors:\n" + \
                                            "\n".join(modInfo.lexerErrors + \
                                                      modInfo.errors)
                            node.parsingErrors = True

                        if modInfo.encoding is None and \
                           not modInfo.imports and \
                           not modInfo.globals and \
                           not modInfo.functions and \
                           not modInfo.classes:
                            node.populated = True
                            node.lazyPopulation = False

                node.needVCSStatus = parentItem.needVCSStatus
                self._addItem(node, parentItem)

            if repopulate:
                self.endInsertRows()

        parentItem.populated = True

        # Request statuses of the populated items. The request must be sent
        # after the items are added, otherwise the status received by the model
        # before the items are populated thus not updated properly.
        for path in pathsToRequest:
            GlobalData().mainWindow.vcsManager.requestStatus(path)
        QApplication.restoreOverrideCursor()

    def populateSysPathItem(self, parentItem, repopulate=False):
        """Populates the sys.path item's subtree"""
        if sys.path:
            if repopulate:
                self.beginInsertRows(self.createIndex(parentItem.row(),
                                                      0, parentItem),
                                     0, len(sys.path) - 1)
            for path in sys.path:
                if path == '':
                    path = os.getcwd()

                if os.path.isdir(path):
                    node = TreeViewDirectoryItem(parentItem, path)
                    self._addItem(node, parentItem)
            if repopulate:
                self.endInsertRows()

        parentItem.populated = True

    def populateFileItem(self, parentItem, repopulate=False):
        """Populate a file item's subtree"""
        path = parentItem.getPath()
        if not isPythonFile(path):
            return

        parentItem.populated = True
        modInfo = self.globalData.briefModinfoCache.get(path)

        # Count the number of rows to insert
        count = 0
        if modInfo.encoding is not None:
            count += 1
        if modInfo.imports:
            count += 1
        if modInfo.globals:
            count += 1
        if modInfo.functions:
            count += 1
        if modInfo.classes:
            count += 1

        if count == 0:
            return

        # Insert rows
        if repopulate:
            self.beginInsertRows(self.createIndex(parentItem.row(),
                                                  0, parentItem),
                                 0, count - 1)
        if modInfo.encoding is not None:
            node = TreeViewCodingItem(parentItem, modInfo.encoding)
            self._addItem(node, parentItem)

        if modInfo.imports:
            node = TreeViewImportsItem(parentItem, modInfo)
            self._addItem(node, parentItem)

        if modInfo.globals:
            node = TreeViewGlobalsItem(parentItem, modInfo)
            self._addItem(node, parentItem)

        if modInfo.functions:
            node = TreeViewFunctionsItem(parentItem, modInfo)
            self._addItem(node, parentItem)

        if modInfo.classes:
            node = TreeViewClassesItem(parentItem, modInfo)
            self._addItem(node, parentItem)

        if repopulate:
            self.endInsertRows()

    def __populateList(self, parentItem, items, itemClass, repopulate=False):
        """Helper for populating lists"""
        parentItem.populated = True

        if repopulate:
            self.beginInsertRows(
                self.createIndex(parentItem.row(), 0, parentItem),
                0, len(items) - 1)
        for item in items:
            treeItem = itemClass(parentItem, item)
            if parentItem.columnCount() > 1:
                # Find out a parent with path set
                path = self.findParentPath(parentItem)
                treeItem.appendData([os.path.basename(path), item.line])
                treeItem.setPath(path)
            self._addItem(treeItem, parentItem)

        if repopulate:
            self.endInsertRows()

    @staticmethod
    def findParentPath(item):
        """Provides the nearest parent path"""
        while item.path is None:
            item = item.parentItem
        return item.path

    def populateGlobalsItem(self, parentItem, repopulate=False):
        """Populates the globals item"""
        self.__populateList(parentItem, parentItem.sourceObj.globals,
                            TreeViewGlobalItem, repopulate)

    def populateImportsItem(self, parentItem, repopulate=False):
        """Populates the imports item"""
        self.__populateList(parentItem, parentItem.sourceObj.imports,
                            TreeViewImportItem, repopulate)

    def populateFunctionsItem(self, parentItem, repopulate=False):
        """Populates functions item"""
        self.__populateList(parentItem, parentItem.sourceObj.functions,
                            TreeViewFunctionItem, repopulate)

    def populateClassesItem(self, parentItem, repopulate=False):
        """Populate classes item"""
        self.__populateList(parentItem, parentItem.sourceObj.classes,
                            TreeViewClassItem, repopulate)

    def populateClassItem(self, parentItem, repopulate):
        """Populates a class item"""
        parentItem.populated = True

        # Count the number of items
        count = len(parentItem.sourceObj.decorators) + \
                len(parentItem.sourceObj.functions)

        if parentItem.sourceObj.classes:
            count += 1
        if parentItem.sourceObj.classAttributes:
            count += 1
        if parentItem.sourceObj.instanceAttributes:
            count += 1

        if count == 0:
            return

        # Insert rows
        if repopulate:
            self.beginInsertRows(
                self.createIndex(parentItem.row(), 0, parentItem),
                0, count - 1)
        for item in parentItem.sourceObj.decorators:
            node = TreeViewDecoratorItem(parentItem, item)
            if parentItem.columnCount() > 1:
                node.appendData([parentItem.data(1), item.line])
                node.setPath(self.findParentPath(parentItem))
            self._addItem(node, parentItem)

        for item in parentItem.sourceObj.functions:
            node = TreeViewFunctionItem(parentItem, item)
            if parentItem.columnCount() > 1:
                node.appendData([parentItem.data(1), item.line])
                node.setPath(self.findParentPath(parentItem))
            self._addItem(node, parentItem)

        if parentItem.sourceObj.classes:
            node = TreeViewClassesItem(parentItem, parentItem.sourceObj)
            if parentItem.columnCount() > 1:
                node.appendData(["", ""])
            self._addItem(node, parentItem)

        if parentItem.sourceObj.classAttributes:
            node = TreeViewStaticAttributesItem(parentItem)
            if parentItem.columnCount() > 1:
                node.appendData(["", ""])
            self._addItem(node, parentItem)

        if parentItem.sourceObj.instanceAttributes:
            node = TreeViewInstanceAttributesItem(parentItem)
            if parentItem.columnCount() > 1:
                node.appendData(["", ""])
            self._addItem(node, parentItem)

        if repopulate:
            self.endInsertRows()

    def populateFunctionItem(self, parentItem, repopulate):
        """Populates a function item"""
        parentItem.populated = True

        # Count the number of items
        count = len(parentItem.sourceObj.decorators)

        if parentItem.sourceObj.functions:
            count += 1
        if parentItem.sourceObj.classes:
            count += 1

        if count == 0:
            return

        # Insert rows
        if repopulate:
            self.beginInsertRows(
                self.createIndex(parentItem.row(), 0, parentItem),
                0, count - 1)

        for item in parentItem.sourceObj.decorators:
            node = TreeViewDecoratorItem(parentItem, item)
            if parentItem.columnCount() > 1:
                node.appendData([parentItem.data(1), item.line])
                node.setPath(self.findParentPath(parentItem))
            self._addItem(node, parentItem)

        if parentItem.sourceObj.functions:
            node = TreeViewFunctionsItem(parentItem, parentItem.sourceObj)
            if parentItem.columnCount() > 1:
                node.appendData(["", ""])
            self._addItem(node, parentItem)

        if parentItem.sourceObj.classes:
            node = TreeViewClassesItem(parentItem, parentItem.sourceObj)
            if parentItem.columnCount() > 1:
                node.appendData(["", ""])
            self._addItem(node, parentItem)

        if repopulate:
            self.endInsertRows()

    def populateStaticAttributesItem(self, parentItem, repopulate):
        """Populates a static attributes item"""
        self.__populateList(parentItem,
                            parentItem.parentItem.sourceObj.classAttributes,
                            TreeViewAttributeItem, repopulate)

    def populateInstanceAttributesItem(self, parentItem, repopulate):
        """Populates an instance attributes item"""
        self.__populateList(parentItem,
                            parentItem.parentItem.sourceObj.instanceAttributes,
                            TreeViewAttributeItem, repopulate)

    def populateImportItem(self, parentItem, repopulate):
        """Populate an import item"""
        self.__populateList(parentItem, parentItem.sourceObj.what,
                            TreeViewWhatItem, repopulate)

    def signalItemUpdated(self, treeItem):
        """Emits a signal that an item is updated"""
        index = self.buildIndex(treeItem.getRowPath())
        self.dataChanged.emit(index, index)

    def removeTreeItem(self, treeItem):
        """Removes the given item"""
        index = self.buildIndex(treeItem.getRowPath())
        self.beginRemoveRows(index.parent(), index.row(), index.row())
        treeItem.parentItem.removeChild(treeItem)
        self.endRemoveRows()

    def addTreeItem(self, treeItem, newItem):
        """Adds the given item"""
        parentIndex = self.buildIndex(treeItem.getRowPath())
        self.addItem(newItem, parentIndex)

    def updateSingleClassItem(self, treeItem, classObj):
        """Updates single class item"""
        if not treeItem.populated:
            return

        # There might be decorators, classes, methods, static attributes and
        # instance attributes
        existingDecors = []
        existingMethods = []
        hadStaticAttributes = False
        hadInstanceAttributes = False
        hadClasses = False
        itemsToRemove = []
        for classChildItem in treeItem.childItems:
            if classChildItem.itemType == DecoratorItemType:
                name = classChildItem.sourceObj.name
                found = False
                for decor in classObj.decorators:
                    if decor.name == name:
                        found = True
                        existingDecors.append(name)
                        if cmpDecoratorDisplayName(classChildItem.sourceObj,
                                                   decor):
                            classChildItem.updateData(decor)
                            classChildItem.setData(2, decor.line)
                        else:
                            # Appearence changed
                            classChildItem.updateData(decor)
                            classChildItem.setData(2, decor.line)
                            self.signalItemUpdated(classChildItem)
                        break
                if not found:
                    itemsToRemove.append(classChildItem)
                continue
            if classChildItem.itemType == ClassesItemType:
                hadClasses = True
                if not classObj.classes:
                    itemsToRemove.append(classChildItem)
                else:
                    classChildItem.updateData(classObj)
                    self.updateClassesItem(classChildItem,
                                           classObj.classes)
                continue
            if classChildItem.itemType == FunctionItemType:
                name = classChildItem.sourceObj.name
                found = False
                for method in classObj.functions:
                    if method.name == name:
                        found = True
                        existingMethods.append(name)
                        if cmpFunctionDisplayName(classChildItem.sourceObj,
                                                  method):
                            classChildItem.updateData(method)
                            classChildItem.setData(2, method.line)
                        else:
                            # Appearence changed
                            classChildItem.updateData(method)
                            classChildItem.setData(2, method.line)
                            self.signalItemUpdated(classChildItem)
                        self.updateSingleFuncItem(classChildItem,
                                                  method)
                        break
                if not found:
                    itemsToRemove.append(classChildItem)
                continue
            if classChildItem.itemType == StaticAttributesItemType:
                hadStaticAttributes = True
                if not classObj.classAttributes:
                    itemsToRemove.append(classChildItem)
                else:
                    self.updateAttrItem(classChildItem,
                                        classObj.classAttributes)
                continue
            if classChildItem.itemType == InstanceAttributesItemType:
                hadInstanceAttributes = True
                if not classObj.instanceAttributes:
                    itemsToRemove.append(classChildItem)
                else:
                    self.updateAttrItem(classChildItem,
                                        classObj.instanceAttributes)
                continue

        for item in itemsToRemove:
            self.removeTreeItem(item)

        # Add those which have been introduced
        for decor in classObj.decorators:
            if decor.name not in existingDecors:
                newItem = TreeViewDecoratorItem(treeItem, decor)
                if treeItem.columnCount() > 1:
                    newItem.appendData([treeItem.data(1), decor.line])
                    newItem.setPath(self.findParentPath(treeItem))
                self.addTreeItem(treeItem, newItem)
        for method in classObj.functions:
            if method.name not in existingMethods:
                newItem = TreeViewFunctionItem(treeItem, method)
                if treeItem.columnCount() > 1:
                    newItem.appendData([treeItem.data(1), method.line])
                    newItem.setPath(self.findParentPath(treeItem))
                self.addTreeItem(treeItem, newItem)

        if not hadClasses and classObj.classes:
            newItem = TreeViewClassesItem(treeItem, classObj)
            if treeItem.columnCount() > 1:
                newItem.appendData(["", ""])
            self.addTreeItem(treeItem, newItem)
        if not hadStaticAttributes and \
           classObj.classAttributes:
            newItem = TreeViewStaticAttributesItem(treeItem)
            if treeItem.columnCount() > 1:
                newItem.appendData(["", ""])
            self.addTreeItem(treeItem, newItem)
        if not hadInstanceAttributes and \
           classObj.instanceAttributes:
            newItem = TreeViewInstanceAttributesItem(treeItem)
            if treeItem.columnCount() > 1:
                newItem.appendData(["", ""])
            self.addTreeItem(treeItem, newItem)

    def updateSingleFuncItem(self, treeItem, funcObj):
        """Updates single function item"""
        if not treeItem.populated:
            return

        # It may have decor, classes and other functions
        existingDecors = []
        hadFunctions = False
        hadClasses = False
        itemsToRemove = []
        for funcChildItem in treeItem.childItems:
            if funcChildItem.itemType == DecoratorItemType:
                name = funcChildItem.sourceObj.name
                found = False
                for decor in funcObj.decorators:
                    if decor.name == name:
                        found = True
                        existingDecors.append(name)
                        if cmpDecoratorDisplayName(funcChildItem.sourceObj,
                                                   decor):
                            funcChildItem.updateData(decor)
                            funcChildItem.setData(2, decor.line)
                        else:
                            # Appearence changed
                            funcChildItem.updateData(decor)
                            funcChildItem.setData(2, decor.line)
                            self.signalItemUpdated(funcChildItem)
                        break
                if not found:
                    itemsToRemove.append(funcChildItem)
                continue
            if funcChildItem.itemType == FunctionsItemType:
                hadFunctions = True
                if not funcObj.functions:
                    itemsToRemove.append(funcChildItem)
                else:
                    funcChildItem.updateData(funcObj)
                    self.updateFunctionsItem(funcChildItem,
                                             funcObj.functions)
                continue
            if funcChildItem.itemType == ClassesItemType:
                hadClasses = True
                if not funcObj.classes:
                    itemsToRemove.append(funcChildItem)
                else:
                    funcChildItem.updateData(funcObj)
                    self.updateClassesItem(funcChildItem,
                                           funcObj.classes)
                continue

        for item in itemsToRemove:
            self.removeTreeItem(item)

        # Add those which have been introduced
        for decor in funcObj.decorators:
            if decor.name not in existingDecors:
                newItem = TreeViewDecoratorItem(treeItem, decor)
                if treeItem.columnCount() > 1:
                    newItem.appendData([treeItem.data(1), decor.line])
                    newItem.setPath(self.findParentPath(treeItem))
                self.addTreeItem(treeItem, newItem)

        if not hadFunctions and funcObj.functions:
            newItem = TreeViewFunctionsItem(treeItem, funcObj)
            if treeItem.columnCount() > 1:
                newItem.appendData(["", ""])
            self.addTreeItem(treeItem, newItem)
        if not hadClasses and funcObj.classes:
            newItem = TreeViewClassesItem(treeItem, funcObj)
            if treeItem.columnCount() > 1:
                newItem.appendData(["", ""])
            self.addTreeItem(treeItem, newItem)

    def updateClassesItem(self, treeItem, classesObj):
        """Updates classes item"""
        if not treeItem.populated:
            return

        existingClasses = []
        itemsToRemove = []
        for classItem in treeItem.childItems:
            name = classItem.sourceObj.name
            found = False
            for cls in classesObj:
                if cls.name == name:
                    found = True
                    existingClasses.append(name)
                    if cmpClassDisplayName(classItem.sourceObj, cls):
                        classItem.updateData(cls)
                        classItem.setData(2, cls.line)
                    else:
                        # Appearence changed
                        classItem.updateData(cls)
                        classItem.setData(2, cls.line)
                        self.signalItemUpdated(classItem)
                    self.updateSingleClassItem(classItem, cls)
                    break
            if not found:
                itemsToRemove.append(classItem)

        for item in itemsToRemove:
            self.removeTreeItem(item)

        # Add those which have been introduced
        for cls in classesObj:
            if cls.name not in existingClasses:
                newItem = TreeViewClassItem(treeItem, cls)
                if treeItem.columnCount() > 1:
                    newItem.appendData([treeItem.data(1), cls.line])
                    newItem.setPath(self.findParentPath(treeItem))
                self.addTreeItem(treeItem, newItem)

    def updateFunctionsItem(self, treeItem, functionsObj):
        """Updates functions item"""
        if not treeItem.populated:
            return

        existingFunctions = []
        itemsToRemove = []
        for functionItem in treeItem.childItems:
            name = functionItem.sourceObj.name
            found = False
            for func in functionsObj:
                if func.name == name:
                    found = True
                    existingFunctions.append(name)
                    if cmpFunctionDisplayName(functionItem.sourceObj,
                                              func):
                        functionItem.updateData(func)
                        functionItem.setData(2, func.line)
                    else:
                        # Appearence changed
                        functionItem.updateData(func)
                        functionItem.setData(2, func.line)
                        self.signalItemUpdated(functionItem)
                    self.updateSingleFuncItem(functionItem, func)
                    break
            if not found:
                itemsToRemove.append(functionItem)

        for item in itemsToRemove:
            self.removeTreeItem(item)

        # Add those which have been introduced
        for func in functionsObj:
            if func.name not in existingFunctions:
                newItem = TreeViewFunctionItem(treeItem, func)
                if treeItem.columnCount() > 1:
                    newItem.appendData([treeItem.data(1), func.line])
                    newItem.setPath(self.findParentPath(treeItem))
                self.addTreeItem(treeItem, newItem)

    def updateAttrItem(self, treeItem, attributesObj):
        """Updates attributes item"""
        if not treeItem.populated:
            return

        existingAttributes = []
        itemsToRemove = []
        for attrItem in treeItem.childItems:
            name = attrItem.data(0)
            found = False
            for attr in attributesObj:
                if attr.name == name:
                    found = True
                    existingAttributes.append(name)
                    attrItem.updateData(attr)
                    attrItem.setData(2, attr.line)
                    # There is no need to send a signal to update the item
                    # because the only name is displayed and it's not
                    # changed.
                    # self.signalItemUpdated( attrItem )
                    break
            if not found:
                itemsToRemove.append(attrItem)

        for item in itemsToRemove:
            self.removeTreeItem(item)

        for attr in attributesObj:
            if attr.name not in existingAttributes:
                newItem = TreeViewAttributeItem(treeItem, attr)
                if treeItem.columnCount() > 1:
                    newItem.appendData([treeItem.data(1), attr.line])
                    newItem.setPath(self.findParentPath(treeItem))
                self.addTreeItem(treeItem, newItem)


def cmpDocstringDisplayName(lhs, rhs):
    """Returns True if the display names are the same"""
    if lhs is None and rhs is None:
        return True
    if lhs is None or rhs is None:
        return False
    return lhs.text == rhs.text


def cmpFunctionDisplayName(lhs, rhs):
    """Returns True if the functions display names are the same"""
    if lhs.name != rhs.name:
        return False
    if lhs.arguments != rhs.arguments:
        return False
    return cmpDocstringDisplayName(lhs.docstring, rhs.docstring)


def cmpDecoratorDisplayName(lhs, rhs):
    """Returns True if the decorators display names are the same"""
    if lhs.name != rhs.name:
        return False
    if lhs.arguments != rhs.arguments:
        return False
    return True


def cmpClassDisplayName(lhs, rhs):
    """Returns True if the classes display names are the same"""
    if lhs.name != rhs.name:
        return False
    if lhs.base != rhs.base:
        return False
    return cmpDocstringDisplayName(lhs.docstring, rhs.docstring)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""imports diagram graphics objects"""

import math
from ui.qt import (QFont, QPen, QColor, QPainterPath, QFontMetrics, QPainter,
                   QImage, QGraphicsRectItem, QGraphicsPathItem, QWidget,
                   QGraphicsView, QToolBar, QHBoxLayout, QAction,
                   QGraphicsItem, QGraphicsTextItem, QApplication,
                   QStyleOptionGraphicsItem, QStyle, Qt, QSize, QPointF,
                   pyqtSignal)
from ui.mainwindowtabwidgetbase import MainWindowTabWidgetBase
from utils.pixmapcache import getIcon
from utils.globals import GlobalData


class ImportsDgmEdgeLabel(QGraphicsTextItem):

    """Connector label"""

    def __init__(self, edge, modObj):
        text = edge.label.replace('\\n', '\n')
        QGraphicsTextItem.__init__(self, text)
        self.__modObj = modObj

        font = QFont("Arial", 10)
        self.setFont(font)

        metric = QFontMetrics(font)
        rec = metric.boundingRect(0, 0, 10000, 10000, Qt.AlignLeft, text)

        self.setPos(edge.labelX - rec.width() / 2,
                    edge.labelY - rec.height() / 2)

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def paint(self, painter, option, widget):
        """Draws the edge text"""
        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected
        QGraphicsTextItem.paint(self, painter, itemOption, widget)

    def mouseDoubleClickEvent(self, event):
        """Open the clicked file as the new one"""
        # Jump to the first import statement
        line = self.__modObj.imports[0].line
        GlobalData().mainWindow.openFile(self.__modObj.refFile, line)


class ImportsDgmDocConn(QGraphicsPathItem):

    """Connection to a docstring note"""

    def __init__(self, edge, modObj):
        QGraphicsPathItem.__init__(self)
        self.__edge = edge
        self.__modObj = modObj

        startPoint = QPointF(edge.points[0][0], edge.points[0][1])
        painterPath = QPainterPath(startPoint)

        index = 1
        while index + 3 <= len(edge.points):
            painterPath.cubicTo(
                edge.points[index][0], edge.points[index][1],
                edge.points[index+1][0], edge.points[index+1][1],
                edge.points[index+2][0], edge.points[index+2][1])
            index = index + 3
        if index + 2 <= len(edge.points):
            painterPath.quadTo(
                edge.points[index+1][0], edge.points[index+1][1],
                edge.points[index+2][0], edge.points[index+2][1])
            index = index + 2

        if index + 1 <= len(edge.points):
            painterPath.lineTo(
                edge.points[index+1][0], edge.points[index+1][1])

        self.setPath(painterPath)

    def paint(self, painter, option, widget):
        """Draws a curve and then adds an arrow"""
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        pen.setStyle(Qt.DotLine)
        self.setPen(pen)

        QGraphicsPathItem.paint(self, painter, option, widget)


class ImportsDgmDependConn(QGraphicsPathItem):

    """Connection to a dependency module"""

    def __init__(self, edge, modObj, connObj):
        QGraphicsPathItem.__init__(self)
        self.__edge = edge
        self.__modObj = modObj
        self.__connObj = connObj

        startPoint = QPointF(edge.points[0][0], edge.points[0][1])
        painterPath = QPainterPath(startPoint)

        index = 1
        while index + 3 <= len(edge.points):
            painterPath.cubicTo(edge.points[index][0], edge.points[index][1],
                                edge.points[index+1][0],edge.points[index+1][1],
                                edge.points[index+2][0],edge.points[index+2][1])
            index = index + 3
        if index + 2 <= len(edge.points):
            painterPath.quadTo(edge.points[index+1][0], edge.points[index+1][1],
                               edge.points[index+2][0], edge.points[index+2][1])
            index = index + 2

        if index + 1 <= len(edge.points):
            painterPath.lineTo(edge.points[index+1][0], edge.points[index+1][1])

        lastIndex = len(edge.points) - 1
        self.addArrow(painterPath,
                      edge.points[lastIndex-1][0], edge.points[lastIndex-1][1],
                      edge.points[lastIndex][0], edge.points[lastIndex][1])
        self.setPath(painterPath)

    def addArrow(self, painterPath, startX, startY, endX, endY):
        """Add arrows to the edges
           http://kapo-cpp.blogspot.com/2008/10/drawing-arrows-with-cairo.html
        """
        arrowLength = 12.0
        arrowDegrees = 0.15      # Radian

        angle = math.atan2(endY - startY, endX - startX) + math.pi
        arrowX1 = endX + arrowLength * math.cos(angle - arrowDegrees)
        arrowY1 = endY + arrowLength * math.sin(angle - arrowDegrees)
        arrowX2 = endX + arrowLength * math.cos(angle + arrowDegrees)
        arrowY2 = endY + arrowLength * math.sin(angle + arrowDegrees)

        painterPath.moveTo(endX, endY)
        painterPath.lineTo(arrowX1, arrowY1)
        painterPath.moveTo(endX, endY)
        painterPath.lineTo(arrowX2, arrowY2)

    def paint(self, painter, option, widget):
        """Draws a curve and then adds an arrow"""
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        self.setPen(pen)

        painter.setRenderHint(QPainter.Antialiasing, True)
        QGraphicsPathItem.paint(self, painter, option, widget)


class ImportsDgmUnknownModule(QGraphicsRectItem):

    """Unknown module"""

    def __init__(self, node):
        QGraphicsRectItem.__init__(self)
        self.__node = node

        posX = node.posX - node.width / 2.0
        posY = node.posY - node.height / 2.0
        QGraphicsRectItem.__init__(self, posX, posY,
                                   node.width, node.height)
        pen = QPen(QColor(153, 0, 0))
        pen.setWidth(2)
        self.setPen(pen)

        self.setBrush(QColor(216, 216, 207))

    def paint(self, painter, option, widget):
        """Draws a filled rectangle and then adds a title"""
        # Draw the rectangle
        QGraphicsRectItem.paint(self, painter, option, widget)

        # Draw text over the rectangle
        font = QFont("Arial", 10)
        font.setBold(True)
        painter.setFont(font)
        painter.setPen(QPen(QColor(90, 90, 88)))
        painter.drawText(self.__node.posX - self.__node.width / 2.0,
                         self.__node.posY - self.__node.height / 2.0,
                         self.__node.width, self.__node.height,
                         Qt.AlignCenter, self.__node.label)


class ImportsDgmBuiltInModule(QGraphicsRectItem):

    """Built-in module"""

    def __init__(self, node):
        QGraphicsRectItem.__init__(self)
        self.__node = node

        posX = node.posX - node.width / 2.0
        posY = node.posY - node.height / 2.0
        QGraphicsRectItem.__init__(self, posX, posY,
                                   node.width, node.height)
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        self.setPen(pen)

        self.setBrush(QColor(216, 216, 207))

    def paint(self, painter, option, widget):
        """Draws a filled rectangle and then adds a title"""
        # Draw the rectangle
        QGraphicsRectItem.paint(self, painter, option, widget)

        # Draw text over the rectangle
        font = QFont("Arial", 10)
        font.setBold(True)
        painter.setFont(font)
        painter.setPen(QPen(QColor(90, 90, 88)))
        painter.drawText(self.__node.posX - self.__node.width / 2.0,
                         self.__node.posY - self.__node.height / 2.0,
                         self.__node.width, self.__node.height,
                         Qt.AlignCenter, self.__node.label)


class ImportsDgmSystemWideModule(QGraphicsRectItem):

    """Systemwide module"""

    def __init__(self, node, refFile, docstring):
        QGraphicsRectItem.__init__(self)
        self.__node = node
        self.__refFile = refFile
        self.__docstring = docstring

        posX = node.posX - node.width / 2.0
        posY = node.posY - node.height / 2.0
        QGraphicsRectItem.__init__(self, posX, posY,
                                   node.width, node.height)
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        self.setPen(pen)

        self.setBrush(QColor(220, 255, 220))

        # System modules are clickable
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def paint(self, painter, option, widget):
        """Draws a filled rectangle and then adds a title"""
        # Hide the dotted outline for clickable system modules
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected

        # Draw the rectangle
        QGraphicsRectItem.paint(self, painter, itemOption, widget)

        # Draw text over the rectangle
        font = QFont("Arial", 10)
        font.setBold(True)
        painter.setFont(font)
        painter.drawText(self.__node.posX - self.__node.width / 2.0,
                         self.__node.posY - self.__node.height / 2.0,
                         self.__node.width, self.__node.height,
                         Qt.AlignCenter, self.__node.label)

    def mouseDoubleClickEvent(self, event):
        """Open the clicked file as the new one"""
        if self.__refFile != "":
            GlobalData().mainWindow.openFile(self.__refFile, -1)


class ImportsDgmDetailedModuleBase(QGraphicsRectItem):

    """Base class which calculates section heights"""

    def __init__(self, node, refFile, srcobj, deviceDPI):
        self.__node = node
        self.__srcobj = srcobj
        self.__refFile = refFile

        self.__pixelsPerLine = self.calcPixelsPerLine()

        posX = node.posX - node.width / 2.0
        posY = node.posY - node.height / 2.0
        QGraphicsRectItem.__init__(self, posX, posY,
                                   node.width, node.height)
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        self.setPen(pen)

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def calcPixelsPerLine(self):
        """Provides the module section heights"""
        self.__lines = [self.__srcobj.title, None]
        for klass in self.__srcobj.classes:
            self.__lines.append(klass.name)
        self.__lines.append(None)
        for func in self.__srcobj.funcs:
            self.__lines.append(func.name)
        self.__lines.append(None)
        for glob in self.__srcobj.globs:
            self.__lines.append(glob.name)

        # One line a spare for half a line at the top and half a line at the
        # bottom
        return int(float(self.__node.height) / float(len(self.__lines) + 1))

    def paint(self, painter, option, widget):
        """Draws a rectangle, adds title, classes/funcs/globs sections"""
        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected

        # Draw the rectangle
        QGraphicsRectItem.paint(self, painter, itemOption, widget)

        # Start from the bottom to automatically take potential spare
        # pixels for the title
        font = QFont("Arial", 10)
        painter.setFont(font)
        posX = self.__node.posX - self.__node.width / 2.0
        posY = self.__node.posY + self.__node.height / 2.0 - 1.5 * self.__pixelsPerLine

        occupiedPixels = 0
        for index in range(len(self.__lines) - 1, 0, -1):
            if self.__lines[index] is None:
                # Draw a separation line
                painter.drawLine(posX + 1, posY + self.__pixelsPerLine / 2.0,
                                 posX + self.__node.width,
                                 posY + self.__pixelsPerLine / 2.0)
            elif self.__lines[index] != "":
                # Draw a text line
                # Sometimes the bottom part of 'g' is not drawn so I add 2
                # spare pixels.
                painter.drawText(int(posX), int(posY),
                                 int(self.__node.width), self.__pixelsPerLine + 2,
                                 Qt.AlignCenter, self.__lines[index])
            occupiedPixels += self.__pixelsPerLine
            posY -= self.__pixelsPerLine

        # Draw the title in bold
        font.setBold(True)
        painter.setFont(font)

        available = self.__node.height - occupiedPixels
        posY = self.__node.posY - self.__node.height / 2.0
        painter.drawText(int(posX), int(posY),
                         int(self.__node.width), int(available),
                         Qt.AlignCenter, self.__lines[0])

    def mouseDoubleClickEvent(self, event):
        """Open the clicked file as the new one"""
        GlobalData().mainWindow.openFile(self.__refFile, -1)


class ImportsDgmModuleOfInterest(ImportsDgmDetailedModuleBase):

    """Module of interest"""

    def __init__(self, node, refFile, srcobj, deviceDPI):
        ImportsDgmDetailedModuleBase.__init__(self, node, refFile,
                                              srcobj, deviceDPI)
        self.setBrush(QColor(224, 236, 255))


class ImportsDgmOtherPrjModule(ImportsDgmDetailedModuleBase):

    """Other in-project module"""

    def __init__(self, node, refFile, srcobj, deviceDPI):
        ImportsDgmDetailedModuleBase.__init__(self, node, refFile,
                                              srcobj, deviceDPI)
        self.setBrush(QColor(240, 240, 110))


class ImportsDgmDocNote(QGraphicsRectItem):

    """Docstring box"""

    def __init__(self, node, refFile, srcobj):
        QGraphicsRectItem.__init__(self)
        self.__node = node
        self.__srcobj = srcobj
        self.__refFile = refFile

        posX = node.posX - node.width / 2.0
        posY = node.posY - node.height / 2.0
        QGraphicsRectItem.__init__(self, posX, posY,
                                   node.width, node.height)
        pen = QPen(QColor(0, 0, 0))
        pen.setWidth(2)
        self.setPen(pen)

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

        self.setBrush(QColor(253, 245, 145))

    def paint(self, painter, option, widget):
        """Draws a filled rectangle and then adds a title"""
        # Hide the dotted outline
        itemOption = QStyleOptionGraphicsItem(option)
        if itemOption.state & QStyle.State_Selected != 0:
            itemOption.state = itemOption.state & ~QStyle.State_Selected

        # Draw the rectangle
        QGraphicsRectItem.paint(self, painter, itemOption, widget)

        # Draw text over the rectangle
        font = QFont("Arial", 10)
        painter.setFont(font)
        hSpacer = 10
        ySpacer = 5
        painter.drawText(self.__node.posX - self.__node.width / 2.0 + hSpacer,
                         self.__node.posY - self.__node.height / 2.0 + ySpacer,
                         self.__node.width - hSpacer,
                         self.__node.height - ySpacer,
                         Qt.AlignLeft, self.__srcobj.text)

    def mouseDoubleClickEvent(self, event):
        """Open the clicked file as the new one"""
        GlobalData().mainWindow.openFile(self.__refFile,
                                         self.__srcobj.line)


class DiagramWidget(QGraphicsView):

    """Widget to show a generated diagram"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):
        QGraphicsView.__init__(self, parent)
        # self.setRenderHint(QPainter.Antialiasing)
        # self.setRenderHint(QPainter.TextAntialiasing)

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            self.sigEscapePressed.emit()
            event.accept()
        elif event.key() == Qt.Key_C and \
             event.modifiers() == Qt.ControlModifier:
            self.onCopy()
            event.accept()
        else:
            QGraphicsView.keyPressEvent(self, event)

    def setScene(self, scene):
        """Sets the scene to display"""
        scene.setBackgroundBrush(GlobalData().skin['nolexerPaper'])
        QGraphicsView.setScene(self, scene)

    def resetZoom(self):
        """Resets the zoom"""
        self.resetTransform()

    def zoomIn(self):
        """Zoom when a button clicked"""
        factor = 1.41 ** (120.0 / 240.0)
        self.scale(factor, factor)

    def zoomOut(self):
        """Zoom when a button clicked"""
        factor = 1.41 ** (-120.0 / 240.0)
        self.scale(factor, factor)

    def wheelEvent(self, event):
        """Mouse wheel event"""
        if QApplication.keyboardModifiers() == Qt.ControlModifier:
            if event.angleDelta().y() < 0:
                self.zoomOut()
            else:
                self.zoomIn()
        else:
            QGraphicsView.wheelEvent(self, event)

    def __getImage(self):
        """Renders the diagram to an image"""
        scene = self.scene()
        image = QImage(scene.width(), scene.height(),
                       QImage.Format_ARGB32_Premultiplied)
        painter = QPainter(image)
        # If switched on then rectangles edges will not be sharp
        # painter.setRenderHint( QPainter.Antialiasing )
        scene.render(painter)
        painter.end()
        return image

    def onCopy(self):
        """Copies the diagram to the exchange buffer"""
        QApplication.clipboard().setImage(self.__getImage())

    def onSaveAs(self, fName):
        """Saves the rendered image to a file"""
        self.__getImage().save(fName, "PNG")


class ImportDgmTabWidget(QWidget, MainWindowTabWidgetBase):

    """Widget for an editors manager"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):
        MainWindowTabWidgetBase.__init__(self)
        QWidget.__init__(self, parent)

        self.__viewer = DiagramWidget(self)
        self.__viewer.sigEscapePressed.connect(self.__onEsc)

        self.__createLayout()

    def __createLayout(self):
        """Creates the toolbar and layout"""
        # Buttons
        printButton = QAction(getIcon('printer.png'), 'Print', self)
        # printButton.setShortcut('Ctrl+')
        printButton.triggered.connect(self.__onPrint)

        printPreviewButton = QAction(
            getIcon('printpreview.png'), 'Print preview', self)
        # printPreviewButton.setShortcut('Ctrl+')
        printPreviewButton.triggered.connect(self.__onPrintPreview)

        fixedSpacer = QWidget()
        fixedSpacer.setFixedHeight(16)

        zoomInButton = QAction(getIcon('zoomin.png'), 'Zoom in (Ctrl+=)', self)
        zoomInButton.setShortcut('Ctrl+=')
        zoomInButton.triggered.connect(self.onZoomIn)

        zoomOutButton = QAction(
            getIcon('zoomout.png'), 'Zoom out (Ctrl+-)', self)
        zoomOutButton.setShortcut('Ctrl+-')
        zoomOutButton.triggered.connect(self.onZoomOut)

        zoomResetButton = QAction(
            getIcon('zoomreset.png'), 'Zoom reset (Ctrl+0)', self)
        zoomResetButton.setShortcut('Ctrl+0')
        zoomResetButton.triggered.connect(self.onZoomReset)

        # Toolbar
        toolbar = QToolBar(self)
        toolbar.setOrientation(Qt.Vertical)
        toolbar.setMovable(False)
        toolbar.setAllowedAreas(Qt.RightToolBarArea)
        toolbar.setIconSize(QSize(16, 16))
        toolbar.setFixedWidth(28)
        toolbar.setContentsMargins(0, 0, 0, 0)
        # toolbar.addAction(printPreviewButton)
        # toolbar.addAction(printButton)
        # toolbar.addWidget(fixedSpacer)
        toolbar.addAction(zoomInButton)
        toolbar.addAction(zoomOutButton)
        toolbar.addAction(zoomResetButton)

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 0, 0, 0)
        hLayout.setSpacing(0)
        hLayout.addWidget(self.__viewer)
        hLayout.addWidget(toolbar)

        self.setLayout(hLayout)

    def setFocus(self):
        """Overridden setFocus"""
        self.__viewer.setFocus()

    def setScene(self, scene):
        """Sets the graphics scene to display"""
        self.__viewer.setScene(scene)

    def __onPrint(self):
        """Triggered on the 'print' button"""
        pass

    def __onPrintPreview(self):
        """Triggered on the 'print preview' button"""
        pass

    def onZoomIn(self):
        """Triggered on the 'zoom in' button"""
        self.__viewer.zoomIn()

    def onZoomOut(self):
        """Triggered on the 'zoom out' button"""
        self.__viewer.zoomOut()

    def onZoomReset(self):
        """Triggered on the 'zoom reset' button"""
        self.__viewer.resetZoom()

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def onCopy(self):
        """Copies the diagram to the exchange buffer"""
        self.__viewer.onCopy()

    def onSaveAs(self, fName):
        """Saves the diagram into the given file"""
        self.__viewer.onSaveAs(fName)

    # Mandatory interface part is below

    def isModified(self):
        """Tells if the file is modified"""
        return False

    def getRWMode(self):
        """Tells if the file is read only"""
        return "RO"

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.GeneratedDiagram

    def getLanguage(self):
        """Tells the content language"""
        return "Diagram"

    def setFileName(self, name):
        """Sets the file name - not applicable"""
        raise Exception("Setting a file name for a diagram is not applicable")

    def setEncoding(self, newEncoding):
        """Sets the new encoding - not applicable for the diagram viewer"""
        return

    def getShortName(self):
        """Tells the display name"""
        return "Imports diagram"

    def setShortName(self, name):
        """Sets the display name - not applicable"""
        raise Exception("Setting a file name for a diagram is not applicable")



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""definition of the codimension QT based application class"""

import qdarkstyle
from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from .garbagecollector import GarbageCollector
from .qt import Qt, QEvent, QApplication, QMenuBar


KEY_PRESS = QEvent.KeyPress
APP_ACTIVATE = QEvent.ApplicationActivate
APP_DEACTIVATE = QEvent.ApplicationDeactivate


class CodimensionApplication(QApplication):

    """Codimension application class"""

    def __init__(self, argv, style):
        QApplication.__init__(self, argv)

        # Sick! The QT doc recommends the following:
        # "To ensure that the application's style is set correctly, it is best
        # to call this function before the QApplication constructor, if
        # possible". However if I do it before QApplication.__init__() then
        # there is a crash! At least with some styles on Ubuntu 12.04 64 bit.
        # So I have to call the initialization after the __init__ call.
        QApplication.setStyle(style)

        self.mainWindow = None
        self.__lastFocus = None
        self.__beforeMenuBar = None

        # Sick! It seems that QT sends Activate/Deactivate signals every time
        # a dialog window is opened/closed. This happens very quickly (and
        # totally unexpected!). So the last focus widget must not be focused
        # unconditionally as the last focus may come from a dialog which has
        # already been destroyed. Without checking that a widget is still alive
        # (e.g. clicking 'Cancel' in a dialog box) leads to a core dump.

        QApplication.setWindowIcon(getIcon('icon.png'))

        self.focusChanged.connect(self.__onFocusChanged)

        appCSS = ''
        if GlobalData().skin['dark']:
            appCSS = qdarkstyle.load_stylesheet(qt_api='pyqt5')
            #appCSS = appCSS.replace('QStatusBar QLabel',
            #                        'QStatusBar QComboBox')
            #appCSS = appCSS.replace('QStatusBar::item',
            #                        'QStatusBar QComboBox')

        # Avoid having rectangular frames on the status bar and
        # some application wide style changes
        adjustmentCSS = GlobalData().skin['appCSS']
        if adjustmentCSS:
            appCSS += '\n' + adjustmentCSS

        if appCSS:
            self.setStyleSheet(appCSS)

        # Install custom GC
        self.__gc = GarbageCollector(self)

        self.installEventFilter(self)

    def setMainWindow(self, window):
        """Memorizes the new window reference"""
        self.mainWindow = window

    @staticmethod
    def __areWidgetsAlive(widget1, widget2):
        """True, True if QT still has the widgets"""
        first = False
        second = False
        for item in QApplication.allWidgets():
            if not first and item == widget1:
                first = True
                if second:
                    return first, second
            if not second and item == widget2:
                second = True
                if first:
                    return first, second
        return first, second

    def eventFilter(self, obj, event):
        """Event filter to catch ESC application wide.

        Pass focus explicitly on broken window managers when the app is
        activated; Catch Ctrl+1 and Ctrl+2 application wide;
        """
        del obj     # unused argument
        try:
            eventType = event.type()
            if eventType == KEY_PRESS:
                key = event.key()
                modifiers = int(event.modifiers())
                if key == Qt.Key_Escape:
                    if self.mainWindow:
                        self.mainWindow.hideTooltips()
                if modifiers == int(Qt.ControlModifier):
                    if key == Qt.Key_1:
                        if self.mainWindow:
                            return self.mainWindow.passFocusToEditor()
                    elif key == Qt.Key_2:
                        if self.mainWindow:
                            return self.mainWindow.passFocusToFlow()
            elif eventType == APP_ACTIVATE:
                lastFocus, \
                beforeMenuBar = self.__areWidgetsAlive(self.__lastFocus,
                                                       self.__beforeMenuBar)
                if lastFocus:
                    if isinstance(self.__lastFocus, QMenuBar):
                        if beforeMenuBar:
                            self.__beforeMenuBar.setFocus()
                self.__lastFocus = None
                if self.mainWindow:
                    self.mainWindow.checkOutsideFileChanges()
            elif eventType == APP_DEACTIVATE:
                if QApplication.activeModalWidget() is not None:
                    self.__lastFocus = None
                else:
                    self.__lastFocus = QApplication.focusWidget()
        except:
            pass
        return False

    def __onFocusChanged(self, fromWidget, toWidget):
        """Triggered when a focus is passed from one widget to another"""
        if isinstance(toWidget, QMenuBar):
            self.__beforeMenuBar = fromWidget
